#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for the MRK EOS models from :cite:t:`HP91,HP98`

These are not intended to be used directly, but rather as a building block of the CORK models they
serve as convenient tests in the absence of a virial correction term."""

from atmodeller.eos import RealGas
from atmodeller.eos._holland_powell import (
    CO2_mrk_cs_holland91,
    CO2MrkHolland91,
    H2O_mrk_fluid_holland91,
    H2O_mrk_gas_holland91,
    H2OMrkHolland91,
)


def test_CO2_corresponding_states(check_values) -> None:
    """CO2 corresponding states"""
    model: RealGas = CO2_mrk_cs_holland91
    expected: float = 1.5831992703027848
    check_values.fugacity_coefficient(2000, 2e3, model, expected)


def test_CO2(check_values) -> None:
    """CO2"""
    model: RealGas = CO2MrkHolland91
    expected: float = 1.575457075165528
    check_values.fugacity_coefficient(2000, 2e3, model, expected)


def test_H2O_above_Tc(check_values) -> None:
    """H2O above Tc"""
    model: RealGas = H2O_mrk_fluid_holland91
    expected: float = 1.048278616058322
    check_values.fugacity_coefficient(2000, 1e3, model, expected)


def test_H2O_below_Tc_below_Psat(check_values) -> None:
    """H2O below Tc and below Psat"""
    # Psat = 0.118224 kbar at T = 600 K
    model: RealGas = H2O_mrk_gas_holland91
    expected: float = 0.7910907770688191
    check_values.fugacity_coefficient(600, 0.1e3, model, expected)


def test_H2O_below_Tc_above_Psat(check_values) -> None:
    """H2O below Tc and above Psat"""
    # Psat = 0.118224 kbar at T = 600 K
    expected: float = 0.13704706029361396
    check_values.fugacity_coefficient(600, 1e3, H2OMrkHolland91, expected)


# def test_volume_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = CO2MrkHolland91
#     check_values.check_broadcasting("volume", model)


# def test_fugacity_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = CO2MrkHolland91
#     check_values.check_broadcasting("fugacity", model)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Utilities for tests"""

import logging

import pytest
from jaxtyping import ArrayLike
from pytest import approx

from atmodeller.interfaces import SolubilityProtocol

logger: logging.Logger = logging.getLogger("atmodeller.tests.solubility")

# Tolerances to compare the test results with target output.
# RTOL: float = 1.0e-8
RTOL: float = 0.61
"""Relative tolerance"""
# ATOL: float = 1.0e-8
ATOL: float = 0.61
"""Absolute tolerance"""


class CheckValues:
    """Helper to check and confirm values"""

    @classmethod
    def concentration(
        cls,
        function_name: str,
        solubility_model: SolubilityProtocol,
        expected_concentration: ArrayLike,
        fugacity: ArrayLike,
        temperature: ArrayLike,
        pressure: ArrayLike,
        fO2: ArrayLike,
        *,
        rtol=RTOL,
        atol=ATOL,
    ) -> None:
        concentration: ArrayLike = solubility_model.concentration(
            fugacity, temperature=temperature, pressure=pressure, fO2=fO2
        )
        logger.debug("%s, concentration = %s ppmw", function_name, concentration)

        assert concentration == approx(expected_concentration, rtol, atol)


@pytest.fixture(scope="module")
def check_values():
    return CheckValues()



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Redox buffers"""

from abc import abstractmethod
from typing import Optional

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike, Bool

from atmodeller import override
from atmodeller.utilities import (
    ExperimentalCalibration,
    all_not_nan,
    as_j64,
    to_native_floats,
    unit_conversion,
)


class RedoxBuffer(eqx.Module):
    """Redox buffer

    This must adhere to FugacityConstraintProtocol

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to zero.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
    """

    log10_shift: Array
    """Log10 shift"""
    evaluation_pressure: Optional[ArrayLike]
    """Evaluation pressure"""

    def __init__(self, log10_shift: ArrayLike = 0, evaluation_pressure: Optional[ArrayLike] = 1):
        self.log10_shift = as_j64(log10_shift)
        self.evaluation_pressure = evaluation_pressure

    @abstractmethod
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Converts the pressure units

        Args:
            pressure: Pressure in bar

        Returns:
            Pressure in units appropriate for the calculation
        """

    @abstractmethod
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log10 fugacity at the buffer

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity at the buffer
        """

    def active(self) -> Bool[Array, ""]:
        """True if the redox buffer is active, otherwise False"""
        return all_not_nan(self.log10_shift)

    def get_scaled_pressure(self, pressure: ArrayLike) -> ArrayLike:
        """Gets the scaled pressure.

        Args:
            pressure: Pressure in bar

        Returns:
            Pressure in units appropriate for the calculation
        """
        if self.evaluation_pressure is not None:
            return self.convert_pressure_units(self.evaluation_pressure)
        else:
            return self.convert_pressure_units(pressure)

    def log10_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        return self.log10_fugacity_buffer(temperature, pressure) + self.log10_shift

    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log fugacity
        """
        return jnp.log(10) * self.log10_fugacity(temperature, pressure)


class IronWustiteBufferHirschmann08(RedoxBuffer):
    """Iron-wustite buffer :cite:p:`OP93,HGD08`

    Experimental calibration values are provided in the abstract of :cite:t:`HGD08`.

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to zero.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
    """

    calibration: ExperimentalCalibration
    """Experimental calibration"""

    def __init__(self, log10_shift: ArrayLike = 0, evaluation_pressure: Optional[ArrayLike] = 1):
        super().__init__(log10_shift, evaluation_pressure)
        self.calibration = ExperimentalCalibration(pressure_max=27.5 * unit_conversion.GPa_to_bar)

    @override
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are bar"""
        return pressure

    @override
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        scaled_pressure: ArrayLike = self.get_scaled_pressure(pressure)
        log10_fugacity_buffer: Array = (
            -0.8853 * jnp.log(temperature)
            - 28776.8 / temperature
            + 14.057
            + 0.055 * (scaled_pressure - 1) / temperature
        )

        return log10_fugacity_buffer


class IronWustiteBufferHirschmann21(RedoxBuffer):
    """Iron-wustite buffer :cite:p:`H21`

    Regarding the calibration, :cite:t:`H21` states that: 'It extrapolates smoothly to higher
    temperature, though not calibrated above 3000 K. Extrapolation to lower temperatures (<1000 K)
    or higher pressures (>100 GPa) is not recommended.'

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to zero.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
    """

    calibration: ExperimentalCalibration
    """Experimental calibration"""
    a: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """a coefficients"""
    b: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """b coefficients"""
    c: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """c coefficients"""
    d: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """d coefficients"""
    e: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """e coefficients"""
    f: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """f coefficients"""
    g: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """g coefficients"""
    h: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """h coefficients"""
    x: tuple[float, ...] = eqx.field(converter=to_native_floats)
    """Coefficients to define the threshold to use the hcp iron formulation"""

    def __init__(self, log10_shift: ArrayLike = 0, evaluation_pressure: Optional[ArrayLike] = 1):
        super().__init__(log10_shift, evaluation_pressure)
        self.calibration = ExperimentalCalibration(
            temperature_min=1000, pressure_max=100 * unit_conversion.GPa_to_bar
        )
        self.a = (6.844864, 1.175691e-1, 1.143873e-3, 0, 0)
        self.b = (5.791364e-4, -2.891434e-4, -2.737171e-7, 0, 0)
        self.c = (-7.971469e-5, 3.198005e-5, 0, 1.059554e-10, 2.014461e-7)
        self.d = (-2.769002e4, 5.285977e2, -2.919275, 0, 0)
        self.e = (8.463095, -3.000307e-3, 7.213445e-5, 0, 0)
        self.f = (1.148738e-3, -9.352312e-5, 5.161592e-7, 0, 0)
        self.g = (-7.448624e-4, -6.329325e-6, 0, -1.407339e-10, 1.830014e-4)
        self.h = (-2.782082e4, 5.285977e2, -8.473231e-1, 0, 0)
        self.x = (-18.64, 0.04359, -5.069e-6)

    @override
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are GPa"""
        return pressure * unit_conversion.bar_to_GPa

    def _evaluate_m(self, pressure: ArrayLike, coefficients: tuple[float, ...]) -> Array:
        """Evaluates an m parameter

        Args:
            pressure: Pressure in GPa
            coefficients: Coefficients

        Return:
            m parameter
        """
        m: Array = (
            coefficients[4] * jnp.power(pressure, 0.5)
            + coefficients[0]
            + coefficients[1] * pressure
            + coefficients[2] * jnp.power(pressure, 2)
            + coefficients[3] * jnp.power(pressure, 3)
        )

        return m

    def _evaluate_fO2(
        self,
        temperature: ArrayLike,
        pressure: ArrayLike,
        coefficients: tuple[tuple[float, ...], ...],
    ) -> Array:
        """Evaluates the fO2

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa
            coefficients: Coefficients

        Returns:
            log10fO2
        """
        log10fO2: Array = (
            self._evaluate_m(pressure, coefficients[0])
            + self._evaluate_m(pressure, coefficients[1]) * temperature
            + self._evaluate_m(pressure, coefficients[2]) * temperature * jnp.log(temperature)
            + self._evaluate_m(pressure, coefficients[3]) / temperature
        )

        return log10fO2

    def _fcc_bcc_iron(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """log10fO2 for fcc and bcc iron

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Return:
            log10fO2 for fcc and bcc iron
        """
        log10fO2: Array = self._evaluate_fO2(
            temperature, pressure, (self.a, self.b, self.c, self.d)
        )

        return log10fO2

    def _hcp_iron(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """log10fO2 for hcp iron

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Return:
            log10fO2 for hcp iron
        """
        log10fO2: Array = self._evaluate_fO2(
            temperature, pressure, (self.e, self.f, self.g, self.h)
        )

        return log10fO2

    def _use_hcp(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Check to use hcp iron formulation for fO2

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Returns:
            True/False whether to use the hcp iron formulation
        """
        threshold: Array = (
            self.x[2] * jnp.power(temperature, 2) + self.x[1] * temperature + self.x[0]
        )

        return jnp.array(pressure) > threshold

    @override
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        scaled_pressure: ArrayLike = self.get_scaled_pressure(pressure)

        def hcp_case() -> Array:
            return self._hcp_iron(temperature, scaled_pressure)

        def fcc_bcc_case() -> Array:
            return self._fcc_bcc_iron(temperature, scaled_pressure)

        buffer_value: Array = jnp.where(
            self._use_hcp(temperature, scaled_pressure), hcp_case(), fcc_bcc_case()
        )

        return buffer_value


class IronWustiteBufferHirschmann(RedoxBuffer):
    """Composite iron-wustite buffer using :cite:t:`OP93,HGD08` and :cite:t:`H21`

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to 0.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
    """

    calibration: ExperimentalCalibration
    """Experimental calibration"""
    low_temperature_buffer: IronWustiteBufferHirschmann08
    """Low temperature buffer"""
    high_temperature_buffer: IronWustiteBufferHirschmann21
    """High temperature buffer"""

    def __init__(self, log10_shift: ArrayLike = 0, evaluation_pressure: Optional[ArrayLike] = 1):
        super().__init__(log10_shift, evaluation_pressure)
        self.calibration = ExperimentalCalibration(pressure_max=100 * unit_conversion.GPa_to_bar)
        self.low_temperature_buffer = IronWustiteBufferHirschmann08(
            self.log10_shift, self.evaluation_pressure
        )
        self.high_temperature_buffer = IronWustiteBufferHirschmann21(
            self.log10_shift, self.evaluation_pressure
        )

    @override
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are bar

        Not used for a composite redox buffer but required by the interface.
        """
        return pressure

    def _use_low_temperature(self, temperature: ArrayLike) -> Array:
        """Check to use the low temperature buffer for fO2

        Args:
            temperature: Temperature

        Returns:
            True/False whether to use the low temperature formulation
        """
        return as_j64(temperature) < self.high_temperature_buffer.calibration.temperature_min

    @override
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Gets the log10 fugacity at the buffer

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity at the buffer
        """

        def low_temperature_case() -> ArrayLike:
            return self.low_temperature_buffer.log10_fugacity_buffer(temperature, pressure)

        def high_temperature_case() -> ArrayLike:
            return self.high_temperature_buffer.log10_fugacity_buffer(temperature, pressure)

        buffer_value: Array = jnp.where(
            self._use_low_temperature(temperature),
            low_temperature_case(),
            high_temperature_case(),
        )

        return buffer_value


IronWustiteBuffer: type[RedoxBuffer] = IronWustiteBufferHirschmann



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for C-H-O systems with stable or unstable condensates"""

import logging

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import debug_logger
from atmodeller.classes import InteriorAtmosphere
from atmodeller.containers import Planet, SpeciesCollection
from atmodeller.interfaces import FugacityConstraintProtocol
from atmodeller.output import Output
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import earth_oceans_to_hydrogen_mass

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

RTOL: float = 1.0e-8
"""Relative tolerance"""
ATOL: float = 1.0e-8
"""Absolute tolerance"""
TOLERANCE: float = 5.0e-2
"""Tolerance of log output to satisfy comparison with FactSage and FastChem"""

species: SpeciesCollection = SpeciesCollection.create(
    ("H2_g", "H2O_g", "CO_g", "CO2_g", "CH4_g", "O2_g", "C_cr")
)
CHO_system: InteriorAtmosphere = InteriorAtmosphere(species)


def test_graphite_stable(helper) -> None:
    """Tests graphite stable with around 50% condensed C mass fraction"""

    planet: Planet = Planet(surface_temperature=873)
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(np.nan)
    }
    oceans: ArrayLike = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    c_kg: ArrayLike = 5 * h_kg
    o_kg: ArrayLike = 2.73159e19
    mass_constraints = {"C": c_kg, "H": h_kg, "O": o_kg}

    CHO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = CHO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    factsage_result: dict[str, float] = {
        "O2_g": 1.27e-25,
        "H2_g": 14.564,
        "CO_g": 0.07276,
        "H2O_g": 4.527,
        "CO2_g": 0.061195,
        "CH4_g": 96.74,
        "C_cr_activity": 1.0,
        "mass_C_cr": 3.54162e20,
    }

    assert helper.isclose(solution, factsage_result, log=True, rtol=TOLERANCE, atol=TOLERANCE)


def test_graphite_unstable(helper) -> None:
    """Tests C-H-O system at IW+0.5 with graphite unstable

    Similar to :cite:p:`BHS22{Table E, row 2}`
    """

    planet: Planet = Planet(surface_temperature=1400)
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {"O2_g": IronWustiteBuffer(0.5)}
    oceans: ArrayLike = 3
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    c_kg: ArrayLike = 1 * h_kg
    mass_constraints = {"C": c_kg, "H": h_kg}

    CHO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = CHO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    factsage_result: dict[str, float] = {
        "O2_g": 4.11e-13,
        "H2_g": 236.98,
        "CO_g": 46.42,
        "H2O_g": 337.16,
        "CO2_g": 30.88,
        "CH4_g": 28.66,
        "C_cr_activity": 0.12202,
        "mass_C_cr": 0.0,
    }

    assert helper.isclose(solution, factsage_result, log=True, rtol=TOLERANCE, atol=TOLERANCE)


def test_water_stable(helper) -> None:
    """Condensed water at 10 bar"""

    species: SpeciesCollection = SpeciesCollection.create(("H2_g", "H2O_g", "O2_g", "H2O_l"))
    planet: Planet = Planet(surface_temperature=411.75)
    interior_atmosphere: InteriorAtmosphere = InteriorAtmosphere(species)

    oceans: float = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    o_kg: float = 1.14375e21
    mass_constraints = {"H": h_kg, "O": o_kg}

    interior_atmosphere.solve(
        planet=planet,
        mass_constraints=mass_constraints,
    )
    output: Output = interior_atmosphere.output
    solution: dict[str, ArrayLike] = output.quick_look()

    factsage_result: dict[str, float] = {
        "H2O_g": 3.3596,
        "H2_g": 6.5604,
        "O2_g": 5.6433e-58,
        "H2O_l_activity": 1.0,
        "mass_H2O_l": 1.247201e21,
    }

    assert helper.isclose(solution, factsage_result, log=True, rtol=TOLERANCE, atol=TOLERANCE)


def test_graphite_water_stable(helper) -> None:
    """Tests C and water in equilibrium at 430 K and 10 bar"""

    species: SpeciesCollection = SpeciesCollection.create(
        ("H2O_g", "H2_g", "O2_g", "CO_g", "CO2_g", "CH4_g", "H2O_l", "C_cr")
    )
    planet: Planet = Planet(surface_temperature=430)
    interior_atmosphere: InteriorAtmosphere = InteriorAtmosphere(species)

    h_kg: float = 3.10e20
    c_kg: float = 1.08e20
    o_kg: float = 2.48298883581636e21
    mass_constraints = {"C": c_kg, "H": h_kg, "O": o_kg}

    interior_atmosphere.solve(
        planet=planet,
        mass_constraints=mass_constraints,
    )
    output: Output = interior_atmosphere.output
    solution: dict[str, ArrayLike] = output.quick_look()

    factsage_result: dict[str, float] = {
        "CH4_g": 0.3241,
        "CO2_g": 4.3064,
        "CO_g": 2.77e-6,
        "C_cr_activity": 1.0,
        "H2O_g": 5.3672,
        "H2O_l_activity": 1.0,
        "H2_g": 0.0023,
        "O2_g": 4.74e-48,
        "mass_C_cr": 8.75101e19,
        "mass_H2O_l": 2.74821e21,
    }

    assert helper.isclose(solution, factsage_result, log=True, rtol=TOLERANCE, atol=TOLERANCE)



'''
Collin Cherubim
June 30, 2025
Main IsoFATE script for coulped model.
'''

import numpy as np
from atmodeller_coupler import *
from constants import *
from isofunks import *
from orbit_params import *
from atmodeller.atmodeller import InteriorAtmosphere
from atmodeller.atmodeller import Species
from atmodeller.atmodeller import SpeciesCollection
# from atmodeller.atmodeller.solubility import get_solubility_models
solubility_models = get_solubility_models()

def isocalc(f_atm, Mp, Mstar, F0, Fp, T, d, time = 5e9, mechanism = 'XUV', rad_evol = True,
N_H = 0, N_He = 0, N_D = 0, N_O = 0, N_C = 0, melt_fraction_override = False,
mu = mu_solar, eps = 0.15, activity = 'medium', flux_model = 'power law', stellar_type = 'M1',
Rp_override = False, t_sat = 5e8, step_fn = False, F_final = 0, t_pms = 0, pms_factor = 1e2,
n_steps = int(1e5), t0 = 1e6, rho_rcb = 1.0, RR = True, thermal = True, 
beta = -1.23, n_atmodeller = int(1e2), save_molecules = False, mantle_iron_dict = False):
    """
    This is a test
    
    Description

    Args:
        a (array): a test array
        b (array): a test array

    Returns:
        array: a test array

    Returns:
        array: a test array
    """

    # '''
    # Computes species abundances in ternary mixture of H, D, and He 
    # via time-integrated numerical simulation of atmospheric escape.
    # Note: species 1 is H, species 2 is He, species 3 is D

    # Inputs:
    #  - f_atm: atmospheric mass fraction(s), must be in the form of an array [ndim]
    #  - Mp: planet mass [kg]
    #  - Mstar: stellar mass [kg]
    #  - F0: initial incident XUV flux [W/m2]
    #  - Fp: incident bolometric flux [W/m2]
    #  - T: planet equilibrium temperature [K]
    #  - d: orbtial distance [m]
    #  - time: total simulation time; scalar [yr]
    #  - mechanism: 'XUV', 'XUV+RR', 'CPML', 'XUV+CPML', 'fix phi subcritical',
    #  'fix phi supercritical', 'phi kill'
    #  - rad_evol: set to False to fix planet radius at core radius [Bool]
    #  - N_x: initial abundance for species x [atoms]
    #  - melt_fraction_override: set a fixed mantle melt fraction. If False, melt fraction is calculated based on Mp and T_surface [False or float]
    #  - mu: average atmospheric particle mass, default to H/He solar comp [kg]
    #  - eps: heat transfer efficiency [ndim]
    #  - activity: for Fxuv_hazmat function (uses semi-empirical data from MUSCLES survey); 
    #  'low' (lower quartile), 'medium' (median), 'high' (upper quartile)
    #  - flux_model: 'power law' for analytic power law, 'phoenix' for Fxuv_hazmat, 'Johnstone' for Fxuv_Johnstone
    #  - stellar_type: 'M1', 'K5', or 'G5' must be specified for flux_model == Fxuv_Johnstone
    #  - Rp_override: enter scalar planet radius value to manually set constant radius, note radius will not evolve [m]
    #  - t_sat: saturation time for Fxuv power law; 5e8 matches semi-empirical MUSCLES data [yr]
    #  - n_steps: number of timesteps. Convergence occurs at 1e6.
    #  - t0: simulation start time [yr]
    #  - rho_rcb: gas density at the RCB in CPML phi equatin [kg/m3]
    #  - RR: toggles radiation-recombination effect (Ly alpha cooling; Murray-Clay et al 2009)  [Bool]
    #  - thermal: toggles planet radius contraction in Lopez/Fortney equations (False removes age term) [Bool]
    #  - beta: exponential in Fxuv function; determines rate of XUV decrease. -1.23 consistent with MUSCLES data
    #  - n_atmodeller: interval of timesteps between each Atmodeller call
    #  - save_molecules: save molecular abundances at each time step [True] or only final abundances [False]
    #  - mantle_iron_dict: allow Fe in mantle to react with O2.['type']="dynamic" to allow only molten mantle Fe to react; 
    #  ['type']="static" to allow all mantle Fe to react; specify ['Fe_mass_fraction']
     
    # Output: Dictionary of 2-D arrays [len(f_atm) x n_steps] with keys,
    #  - 'time': simulation time array [s]
    #  - 'Rp': total planet radius [m]
    #  - 'Ratm': convective atm depth [m]
    #  - 'Matm': atmospheric mass [kg]
    #  - 'Vpot': gravitational potential at outer layer [J/kg]
    #  - 'fatm': total atmospheric mass fraction [ndim]
    #  - 'Mloss': atm mass loss per time step [kg]
    #  - 'phi': atm mass flux [kg/m2/s]
    #  - 'phic': critical mass flux for species 2 escape [kg/m2/s]
    #  - 'N_H': H number [atoms]
    #  - 'N_He': He number [atoms]
    #  - 'N_D': D number [atoms]
    #  - 'x1': H molar concentration [ndim]
    #  - 'x2': He molar concentration [ndim]
    #  - 'Phi_H': H number flux [atoms/s/m2]
    #  - 'Phi_He': He number flux [atoms/s/m2]
    #  - 'Phi_D': D number flux [atoms/s/m2]
    # '''
    
###_____Initialize physical values_____###

    b = 1.04e20*T**0.732 # [molecules/m/s] from Mason & Marrero 1970 for H in He
    radius_core = R_core(Mp) # [m]
    R_B = R_Bondi(Mp, mu, T) # Bondi radius [m]
    R_H = R_Hill(Mp, Mstar, d) # Hill radius [m]

###_____Initialize timesteps_____###

    n_tot = n_steps # timesteps
    t0 = t0/s2yr # simulation start time [s]
    t = time/s2yr - t0 # total simulation time [s]
    delta_t = t/n_tot # timestep [s]

###_____Set initial values____###

    ### atmodeller interior
    N_H_int = 0
    N_He_int = 0
    N_D_int = 0
    N_O_int = 0
    N_C_int = 0
    if n_atmodeller == 0:
        T_surf_analytic = 0
        T_surf_atmod = 0
    if N_H != 0 and N_D != 0: # needed to allow D and H to outgas from mantle
        X_DH = N_D/(N_H + N_D) # ignores D in mantle

    H2O_g = Species.create_gas("H2O_g", solubility=solubility_models["H2O_basalt_dixon95"])
    H2_g = Species.create_gas("H2_g", solubility=solubility_models["H2_basalt_hirschmann12"])
    O2_g = Species.create_gas("O2_g")
    CO_g = Species.create_gas("CO_g", solubility=solubility_models["CO_basalt_yoshioka19"])
    CO2_g = Species.create_gas("CO2_g", solubility=solubility_models["CO2_basalt_dixon95"])
    CH4_g = Species.create_gas("CH4_g", solubility=solubility_models["CH4_basalt_ardia13"])
    He_g = Species.create_gas("He_g", solubility=solubility_models["He_basalt_jambon86"])

    species = SpeciesCollection((H2_g, H2O_g, O2_g, CO_g, CO2_g, CH4_g, He_g))

    interior_atmosphere = InteriorAtmosphere(species)

    atmod_full_output = {}
    if mantle_iron_dict:
        mantle_iron_dict['mantle_mass'] = 0.704665308539034*Mp # fraction from atmodeller
        mantle_iron_dict['mass_Fe'] = mantle_iron_dict['mantle_mass']*mantle_iron_dict['Fe_mass_fraction']
        mantle_iron_dict['mass_Fe2'] = mantle_iron_dict['mass_Fe']
        mantle_iron_dict['X_Fe2'] = mantle_iron_dict['mass_Fe2']/mantle_iron_dict['mass_Fe']

    ### atmosphere
    M_atm0 = Mp*f_atm # initial atmospheric mass [kg]
    M_atm = M_atm0
    y1 = 1*N_H # H number [atoms]
    y2 = 1*N_He # He number [atoms]
    y3 = 1*N_D # D number [atoms]
    y4 = 1*N_O # O number [atoms]
    y5 = 1*N_C # O number [atoms]

###_____Initialize arrays_____###

    t_a = delta_t*np.linspace(1, n_tot + 1, n_tot) + t0 # time array [s]
    phi_a = np.zeros(n_tot) # mass flux array [kg/s/m2]
    phic_a = np.zeros(n_tot) # critical mass flux array [kg/s/m2] 
    Rp_a = np.zeros(n_tot) # total radius, diagnostic [m]
    Renv_a = np.zeros(n_tot) # envelope radius [m]
    Matm_a = np.zeros(n_tot) # atmospheric mass [kg]
    fatm_a = np.zeros(n_tot) # atm mass fraction [ndim]
    Vpot_a = np.zeros(n_tot) # grav potential, diagnostic [J/kg]
    Mloss_a = np.zeros(n_tot) # mass lost per timestep [kg]

    y1_a = np.zeros(n_tot) # H number array [atoms]
    y2_a = np.zeros(n_tot) # He number array [atoms]
    y3_a = np.zeros(n_tot) # D number array [atoms]
    y4_a = np.zeros(n_tot) # O number array [atoms]
    y5_a = np.zeros(n_tot) # C number array [atoms]
    y1_a_int = np.zeros(n_tot) # mantle H number array [atoms]
    y2_a_int = np.zeros(n_tot) # mantle He number array [atoms]
    y3_a_int = np.zeros(n_tot) # mantle D number array [atoms]
    y4_a_int = np.zeros(n_tot) # mantle O number array [atoms]
    y5_a_int = np.zeros(n_tot) # mantle C number array [atoms]
    H2_a = np.zeros(n_tot) # atmospheric H2 number array [molecules]
    H2O_a = np.zeros(n_tot) # atmospheric H2O number array [molecules]
    O2_a = np.zeros(n_tot) # atmospheric O2 number array [molecules]
    CO2_a = np.zeros(n_tot) # atmospheric CO2 number array [molecules]
    CO_a = np.zeros(n_tot) # atmospheric CO number array [molecules]
    CH4_a = np.zeros(n_tot) # atmospheric CH4 number array [molecules]
    fO2_a = np.zeros(n_tot) # fugacity array [bar]
    x1_a = np.zeros(n_tot) # H molar concentration array [ndim]
    x2_a = np.zeros(n_tot) # He molar concentration array [ndim]
    x3_a = np.zeros(n_tot) # D molar concentration array [ndim]
    x4_a = np.zeros(n_tot) # O molar concentration array [ndim]
    x5_a = np.zeros(n_tot) # C molar concentration array [ndim]
    Phi1_a = np.zeros(n_tot) # H number flux array [atoms/s/m2]
    Phi2_a = np.zeros(n_tot) # He number flux array [atoms/s/m2]
    Phi3_a = np.zeros(n_tot) # D number flux array [atoms/s/m2]
    Phi4_a = np.zeros(n_tot) # O number flux array [atoms/s/m2]
    Phi5_a = np.zeros(n_tot) # C number flux array [atoms/s/m2]
    T_surf_analytic_a = np.zeros(n_tot) # surface temperature from analytic calculation array [K]
    T_surf_atmod_a = np.zeros(n_tot) # atmodeller surface temperature array (capped at 6000 K) [K]

    ###_____Loop through timesteps_____###

    for n in range(n_tot):

    ### Stop simulation when entire atmosphere is lost
        if M_atm <= 0 or y1 + y2 + y3 + y4 + y5 <= 0:
            Matm_a[n:] = 0 #M_atm #Matm_a[n-1]
            fatm_a[n:] = 0 #f_atm #fatm_a[n-1]
            Renv_a[n:] = 0 #radius_env #Renv_a[n-1]
            Rp_a[n:] = radius_core
            K = np.max([V_reduction(Mp, Mstar, d, radius_core), 0.01]) # grav potential reduction factor due to stellar tidal forces
            Vpot_a[n:] = K*G*Mp/radius_core
            phi_a[n:] = 0
            Mloss_a[n:] = 0

            y1_a[n:] = 0 #y1 #y1_a[n-1]
            y2_a[n:] = 0 #y2 #y2_a[n-1]
            y3_a[n:] = 0 #y3 #y3_a[n-1]
            y4_a[n:] = 0 #y4 #y4_a[n-1]
            y5_a[n:] = 0 #y5 #y5_a[n-1]
            x1_a[n:] = x1_a[n-1] # x1_a[max(np.nonzero(x1_a)[0])] # get last non-zero value in array
            x2_a[n:] = x2_a[n-1] # x2_a[max(np.nonzero(x2_a)[0])]
            x3_a[n:] = x3_a[n-1] # x3_a[max(np.nonzero(x3_a)[0])]
            x4_a[n:] = x4_a[n-1] # x4_a[max(np.nonzero(x4_a)[0])]
            x5_a[n:] = x5_a[n-1] # x5_a[max(np.nonzero(x5_a)[0])]
            Phi1_a[n:] = 0
            Phi2_a[n:] = 0    
            Phi3_a[n:] = 0
            Phi4_a[n:] = 0
            Phi5_a[n:] = 0

            # atmodeller full ouput for monte carlo runs
            if n_atmodeller != 0:
                    # atmod_full_output = {}
                atmod_full_output['H2O_atm'] = np.nan
                atmod_full_output['H2O_mantle'] = np.nan
                atmod_full_output['H2_atm'] = np.nan
                atmod_full_output['H2_mantle'] = np.nan
                atmod_full_output['O2_atm'] = np.nan
                atmod_full_output['O2_mantle'] = np.nan
                atmod_full_output['CO_atm'] = np.nan
                atmod_full_output['CO_mantle'] = np.nan
                atmod_full_output['CO2_atm'] = np.nan
                atmod_full_output['CO2_mantle'] = np.nan
                atmod_full_output['CH4_atm'] = np.nan
                atmod_full_output['CH4_mantle'] = np.nan
                atmod_full_output['He_mantle'] = np.nan
                atmod_full_output['O2_fugacity'] = np.nan
                if save_molecules == True:
                    H2_a[n:] = 0
                    H2O_a[n:] = 0
                    O2_a[n:] = 0
                    CO2_a[n:] = 0
                    CO_a[n:] = 0
                    CH4_a[n:] = 0
                    fO2_a[n:] = 0

            break

        # time-variable average atomic mass
        N_tot = y1 + y2 + y3 + y4 + y5
        mu = (y1*mu_H + y2*mu_He + y3*mu_D + y4*mu_O + y5*mu_C)/N_tot

        if rad_evol == False:
            radius_env = 0
            radius_atm = 0
            radius_p = radius_core
            if Rp_override != False:
                radius_core = Rp_override
                radius_env = 0
                radius_atm = 0
        else:
            radius_env = R_env(Mp, f_atm, Fp, t_a[n], thermal)
            radius_atm = R_atm(T, Mp, radius_core, radius_env, mu)
            radius_p = radius_core + radius_atm + radius_env
            radius_p = np.min([R_B, R_H, radius_p]) # limits Rp to the min of Bondi/Hill/Lopez+Fortney radius

        K = np.max([V_reduction(Mp, Mstar, d, radius_p), 0.01]) # grav potential reduction factor due to stellar tidal forces
        Vpot = K*G*Mp/radius_p
        A = 4*np.pi*radius_p**2

    # sets mass flux [kg/m2/s]
        if mechanism == 'XUV':
            if RR == True:
                phi = np.min([phi_RR(radius_p, Mp, T, t_a[n], F0, t0*s2yr, t_sat, beta, step_fn, F_final, t_pms, pms_factor), 
                              phi_E(t_a[n], eps, Vpot, d, F0, t0*s2yr, t_sat, beta, activity, flux_model, stellar_type, step_fn, F_final, t_pms, pms_factor)])
            else:
                phi = phi_E(t_a[n], eps, Vpot, d, F0, t0*s2yr, t_sat, beta, activity, flux_model, stellar_type, step_fn, F_final, t_pms, pms_factor)
        elif mechanism == 'CPML':
            phi = phiE_CP(T, Mp, rho_rcb, eps, Vpot, A, mu, radius_env)
        elif mechanism == 'phi kill':
            phi = phi_kill(Mp*f_atm, radius_p, t - t_a[n])
        elif mechanism == 'XUV+CPML':
            if RR == True:
                phi_XUV = np.min([phi_RR(radius_p, Mp, T, t_a[n], F0, t0*s2yr, t_sat, beta, step_fn, F_final, t_pms, pms_factor), 
                                  phi_E(t_a[n], eps, Vpot, d, F0, t0*s2yr, t_sat, beta, activity, flux_model, stellar_type, step_fn, F_final, t_pms, pms_factor)])
            else:
                phi_XUV = phi_E(t_a[n], eps, Vpot, d, F0, t0*s2yr, t_sat, beta, activity, flux_model, stellar_type, step_fn, F_final, t_pms, pms_factor)
            phi = phi_XUV + phiE_CP(T, Mp, rho_rcb, eps, Vpot, A, mu, radius_env)
        
        mass_loss = phi*A*delta_t
        g = G*Mp/radius_p**2
        H_H = R_gas*T/(M_H*g) # H scale height [m]
        H_He = R_gas*T/(M_He*g) # He scale height [m]
        H_D = R_gas*T/(M_D*g) # D scale height [m]
        H_O = R_gas*T/(M_O*g) # O scale height [m]
        H_C = R_gas*T/(M_C*g) # O scale height [m]

        x1 = y1/N_tot
        x2 = y2/N_tot
        x3 = y3/N_tot
        x4 = y4/N_tot
        x5 = y5/N_tot

        if y1 + y2 == 0:
            X1 = 0
            X2 = 0
        else:
            X1 = y1/(y1+y2)
            X2 = y2/(y1+y2)
        MU = X1*mu_H + X2*mu_He

        Phi1, phi_c = Phi_1(phi, b, H_H, H_He, mu_H, mu_He, X1, X2, MU, output = 1) # H number flux [atoms/s/m2]
        Phi2 = Phi_2(phi, b, H_H, H_He, mu_H, mu_He, X1, X2, MU) # He number flux [atoms/s/m2]
        Phi3 = Phi_D_Z90(Phi1, Phi2, H_H, H_D, H_He, y1, y2, y3, y4, y5, T) # D number flux [atoms/s/m2]
        Phi4 = Phi_O_Z90(Phi1, Phi2, H_H, H_O, H_He, y1, y2, y3, y4, y5, T) # O number flux [atoms/s/m2]
        Phi5 = Phi_C_Z90(Phi1, Phi2, H_H, H_C, H_He, y1, y2, y3, y4, y5, T) # C number flux [atoms/s/m2]

        # record values
        Matm_a[n] = M_atm
        fatm_a[n] = f_atm
        Renv_a[n] = radius_env # this will still change even with Rp limited to min(R_Bondi, R_Hill)
        Rp_a[n] = radius_p
        Vpot_a[n] = Vpot
        phi_a[n] = phi
        phic_a[n] = phi_c
        Mloss_a[n] = mass_loss
        
        y1_a[n] = y1
        y2_a[n] = y2
        y3_a[n] = y3
        y4_a[n] = y4
        y5_a[n] = y5
        y1_a_int[n] = N_H_int
        y2_a_int[n] = N_He_int
        y3_a_int[n] = N_D_int
        y4_a_int[n] = N_O_int
        y5_a_int[n] = N_C_int
        x1_a[n] = x1
        x2_a[n] = x2
        x3_a[n] = x3
        x4_a[n] = x4
        x5_a[n] = x5
        Phi1_a[n] = Phi1
        Phi2_a[n] = Phi2
        Phi3_a[n] = Phi3
        Phi4_a[n] = Phi4
        Phi5_a[n] = Phi5

        ##### run atmodeller ######
        if n_atmodeller != 0: # save final molecular abundances on last time step
            if n == n_steps - 1:
                # atmod_full_output = {}
                atmod_sol = AtmodellerCoupler(T, Mp, radius_p, mu, melt_fraction_override, mantle_iron_dict,
                                                  y1+y3, y2, y4, y5, N_H_int+N_D_int, N_He_int, N_O_int, N_C_int, interior_atmosphere)[1]
                atmod_full_output['H2O_atm'] = atmod_sol['H2O_g']['atmosphere_moles'][0]
                atmod_full_output['H2O_mantle'] = atmod_sol['H2O_g']['dissolved_moles'][0]
                atmod_full_output['H2_atm'] = atmod_sol['H2_g']['atmosphere_moles'][0]
                atmod_full_output['H2_mantle'] = atmod_sol['H2_g']['dissolved_moles'][0]
                atmod_full_output['O2_atm'] = atmod_sol['O2_g']['atmosphere_moles'][0]
                atmod_full_output['O2_mantle'] = atmod_sol['O2_g']['dissolved_moles'][0]
                atmod_full_output['CO_atm'] = atmod_sol['CO_g']['atmosphere_moles'][0]
                atmod_full_output['CO_mantle'] = atmod_sol['CO_g']['dissolved_moles'][0]
                atmod_full_output['CO2_atm'] = atmod_sol['CO2_g']['atmosphere_moles'][0]
                atmod_full_output['CO2_mantle'] = atmod_sol['CO2_g']['dissolved_moles'][0]
                atmod_full_output['CH4_atm'] = atmod_sol['CH4_g']['atmosphere_moles'][0]
                atmod_full_output['CH4_mantle'] = atmod_sol['CH4_g']['dissolved_moles'][0]
                atmod_full_output['He_mantle'] = atmod_sol['He_g']['dissolved_moles'][0]
                atmod_full_output['O2_fugacity'] = atmod_sol['O2_g']['fugacity'][0]
                atmod_full_output['log10dIW_1_bar'] = atmod_sol['O2_g']['log10dIW_1_bar'][0]
            if n%n_atmodeller == 0: # run atmodeller every n_atmodeller steps.
                atmod_results, atmod_full, mantle_iron_dict = AtmodellerCoupler(T, Mp, radius_p, mu, melt_fraction_override, mantle_iron_dict,
                                                  y1+y3, y2, y4, y5, N_H_int+N_D_int, N_He_int, N_O_int, N_C_int, interior_atmosphere)
                N_H_int = atmod_results['N_H_int']*(1 - X_DH)
                N_D_int = atmod_results['N_H_int']*X_DH
                N_He_int = atmod_results['N_He_int']
                N_O_int = atmod_results['N_O_int']
                N_C_int = atmod_results['N_C_int']
                if atmod_results['N_H_atm'] == 0:
                    y1 = 0
                    y3 = 0
                else:
                    Y3 = X_DH*atmod_results['N_H_atm']
                    Y1 = (1 - X_DH)*atmod_results['N_H_atm']
                    y1 = Y1
                    y3 = Y3
                y2 = atmod_results['N_He_atm']
                y4 = atmod_results['N_O_atm']
                y5 = atmod_results['N_C_atm']
                M_atm = atmod_results['M_atm']
                T_surf_analytic = atmod_results['T_surface']
                T_surf_atmod = atmod_results['T_surface_atmod']
                if save_molecules == True:
                    H2_a[n] = atmod_full['H2_g']['atmosphere_moles'][0]
                    H2O_a[n] = atmod_full['H2O_g']['atmosphere_moles'][0]
                    O2_a[n] = atmod_full['O2_g']['atmosphere_moles'][0]
                    CO2_a[n] = atmod_full['CO2_g']['atmosphere_moles'][0]
                    CO_a[n] = atmod_full['CO_g']['atmosphere_moles'][0]
                    CH4_a[n] = atmod_full['CH4_g']['atmosphere_moles'][0]
                    fO2_a[n] = atmod_full['O2_g']['fugacity'][0]
            else:
                H2_a[n] = H2_a[n-1]
                H2O_a[n] = H2O_a[n-1]
                O2_a[n] = O2_a[n-1]
                CO2_a[n] = CO2_a[n-1]
                CO_a[n] = CO_a[n-1]
                CH4_a[n] = CH4_a[n-1]
                fO2_a[n] = fO2_a[n-1]
        T_surf_analytic_a[n] = T_surf_analytic
        T_surf_atmod_a[n] = T_surf_atmod

        if y1 + y3 + N_H_int + N_D_int != 0: # needed to allow D and H to outgas from mantle
            X_DH = (y3 + N_D_int)/(y1 + y3 + N_H_int + N_D_int) # assumes D/H is in equilibrium between interior and atmosphere

        # advance to next step
        y1_loss = Phi1*A*delta_t
        y2_loss = Phi2*A*delta_t
        y3_loss = Phi3*A*delta_t
        y4_loss = Phi4*A*delta_t
        y5_loss = Phi5*A*delta_t
        # M_atm -= mass_loss # comes from phi*A*delta_t
        M_atm -= (y1_loss*mu_H + y2_loss*mu_He + y3_loss*mu_D + y4_loss*mu_O + y5_loss*mu_C)
        f_atm = M_atm/Mp
        y1 -= y1_loss
        y2 -= y2_loss
        y3 -= y3_loss
        y4 -= y4_loss
        y5 -= y5_loss

        y1 = max(y1, 0)
        y2 = max(y2, 0)
        y3 = max(y3, 0)
        y4 = max(y4, 0)
        y5 = max(y5, 0)
    

    # save results
    solutions = {
    'time': t_a,
    'Rp': Rp_a,
    'Ratm': Renv_a,
    'Matm': Matm_a,
    'Vpot': Vpot_a,
    'fatm': fatm_a,
    'Mloss': Mloss_a,
    'phi': phi_a,
    'phic': phic_a,
    'N_H': y1_a,
    'N_He': y2_a,
    'N_D': y3_a,
    'N_O': y4_a,
    'N_C': y5_a,
    'N_H_int': y1_a_int,
    'N_He_int': y2_a_int,
    'N_D_int': y3_a_int,
    'N_O_int': y4_a_int,
    'N_C_int': y5_a_int,
    'x1': x1_a,
    'x2': x2_a,
    'x3': x3_a,
    'x4': x4_a,
    'x5': x5_a,
    'Phi_H': Phi1_a,
    'Phi_He': Phi2_a,
    'Phi_D': Phi3_a,
    'Phi_O': Phi4_a,
    'Phi_C': Phi5_a,
    'T_surf_analytic': T_surf_analytic_a,
    'T_surf_atmod': T_surf_atmod_a
    }
    if save_molecules == True:
        solutions['n_H2_a'] = H2_a,
        solutions['n_H2O_a'] = H2O_a,
        solutions['n_O2_a'] = O2_a,
        solutions['n_CO2_a'] = CO2_a,
        solutions['n_CO_a'] = CO_a,
        solutions['n_CH4_a'] = CH4_a,
        solutions['fO2_a'] = fO2_a
    if n_atmodeller != 0:
        solutions['atmodeller_final'] = atmod_full_output

    return solutions



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility laws for sulfur species

For every law there should be a test in the test suite.
"""

import logging

import jax.numpy as jnp
from jaxtyping import Array, ArrayLike

from atmodeller import override
from atmodeller.solubility.core import Solubility
from atmodeller.utilities import as_j64, unit_conversion

logger: logging.Logger = logging.getLogger(__name__)


class _S2_sulfate_andesite_boulliung23(Solubility):
    """Sulfur as sulfate SO4^2-/S^6+ in andesite :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
    sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for andesite from
    :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
    equilibrated with Air/SO2 mixtures.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (31586.2393 / as_j64(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_andesite_boulliung23: Solubility = _S2_sulfate_andesite_boulliung23()
"""Sulfur as sulfate SO4^2-/S^6+ in andesite :cite:p:`BW22,BW23corr`

Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for andesite from
:cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
equilibrated with Air/SO2 mixtures.
"""


class _S2_sulfide_andesite_boulliung23(Solubility):
    """Sulfur as sulfide (S^2-) in andesite :cite:p:`BW23`

    Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
    for andesite from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
    controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (8921.0927 / as_j64(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_andesite_boulliung23: Solubility = _S2_sulfide_andesite_boulliung23()
"""Sulfur as sulfide (S^2-) in andesite :cite:p:`BW23`

Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
for andesite from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
"""


class _S2_andesite_boulliung23(Solubility):
    """S2 in andesite accounting for both sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility
    _sulfate: Solubility

    def __init__(self):
        self._sulfide = S2_sulfide_andesite_boulliung23
        self._sulfate = S2_sulfate_andesite_boulliung23

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_andesite_boulliung23: Solubility = _S2_andesite_boulliung23()
"""S2 in andesite accounting for both sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""


class _S2_sulfate_basalt_boulliung23(Solubility):
    """Sulfur in basalt as sulfate, SO4^2-/S^6+ :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract and the corrected expression for sulfate capacity
    (C_S6+) in :cite:t:`BW23corr`. Composition for Basalt from :cite:t:`BW22{Table 1}`. Experiments
    conducted at 1 atm pressure, temperatures from 1473-1773 K for silicate melts equilibrated with
    Air/SO2 mixtures.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (32333.5635 / as_j64(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_basalt_boulliung23: Solubility = _S2_sulfate_basalt_boulliung23()
"""Sulfur in basalt as sulfate, SO4^2-/S^6+ :cite:p:`BW22,BW23corr`

Using the first equation in the abstract and the corrected expression for sulfate capacity
(C_S6+) in :cite:t:`BW23corr`. Composition for Basalt from :cite:t:`BW22{Table 1}`. Experiments
conducted at 1 atm pressure, temperatures from 1473-1773 K for silicate melts equilibrated with
Air/SO2 mixtures.
"""


class _S2_sulfide_basalt_boulliung23(Solubility):
    """Sulfur in basalt as sulfide (S^2-) :cite:p:`BW23`

    Using expressions in the abstract for S wt% and sulfide capacity (C_S2-). Composition for
    basalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm pressure and temperatures
    from 1473-1773 K in a controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log
    unit below FMQ.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (8045.7465 / as_j64(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_basalt_boulliung23: Solubility = _S2_sulfide_basalt_boulliung23()
"""Sulfur in basalt as sulfide (S^2-) :cite:p:`BW23`

Using expressions in the abstract for S wt% and sulfide capacity (C_S2-). Composition for
basalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm pressure and temperatures
from 1473-1773 K in a controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log
unit below FMQ.
"""


class _S2_basalt_boulliung23(Solubility):
    """Sulfur in basalt due to sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility
    _sulfate: Solubility

    def __init__(self):
        self._sulfide = S2_sulfide_basalt_boulliung23
        self._sulfate = S2_sulfate_basalt_boulliung23

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_basalt_boulliung23: Solubility = _S2_basalt_boulliung23()
"""Sulfur in basalt due to sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""


class _S2_sulfate_trachybasalt_boulliung23(Solubility):
    """Sulfur as sulfate SO4^2-/S^6+ in trachybasalt :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
    sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for trachybasalt from
    :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
    equilibrated with Air/SO2 mixtures.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (32446.366 / as_j64(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_trachybasalt_boulliung23: Solubility = _S2_sulfate_trachybasalt_boulliung23()
"""Sulfur as sulfate SO4^2-/S^6+ in trachybasalt :cite:p:`BW22,BW23corr`

Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for trachybasalt from
:cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
equilibrated with Air/SO2 mixtures.
"""


class _S2_sulfide_trachybasalt_boulliung23(Solubility):
    """Sulfur as sulfide (S^2-) in trachybasalt :cite:p:`BW23`

    Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
    for trachybasalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
    controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
    """

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (7842.5 / as_j64(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_trachybasalt_boulliung23: Solubility = _S2_sulfide_trachybasalt_boulliung23()
"""Sulfur as sulfide (S^2-) in trachybasalt :cite:p:`BW23`

Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
for trachybasalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
"""


class _S2_trachybasalt_boulliung23(Solubility):
    """Sulfur in trachybasalt by sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility
    _sulfate: Solubility

    def __init__(self):
        self._sulfide = S2_sulfide_trachybasalt_boulliung23
        self._sulfate = S2_sulfate_trachybasalt_boulliung23

    @override
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_trachybasalt_boulliung23: Solubility = _S2_trachybasalt_boulliung23()
"""Sulfur in trachybasalt by sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Package level variables and initialises the package logger"""

__version__: str = "0.9.1"

import logging
import os

import jax
import jax.numpy as jnp
import numpy as np

# from beartype.claw import beartype_this_package
# beartype_this_package()

try:
    from typing import override as _override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override as _override  # Python 3.11 and earlier

override = _override


jax.config.update("jax_enable_x64", True)
jnp.set_printoptions(precision=15)  # For better clarity in printed output
print("Atmodeller initialized with double precision (float64)")

# For debugging
# jax.config.update("jax_debug_nans", True)
# jax.config.update("jax_debug_infs", True)
# jax.config.update("jax_disable_jit", True)
# jax.config.update("jax_log_compiles", True)

# This prevents error_if from throwing an error when encountering NaN or Inf values. To actually
# find the root cause of NaN or Inf values, you should set this to "raise" or "breakpoint" as per
# https://docs.kidger.site/equinox/api/errors/
os.environ["EQX_ON_ERROR"] = "raise"

# Suppress warnings (notably from Equinox about static JAX arrays)
# if not sys.warnoptions:
#     import warnings

#     warnings.simplefilter("ignore")

# Thermodynamic standard state
TEMPERATURE_REFERENCE: float = 298.15
"""Enthalpy reference temperature in K (:math:`T_r` in the JANAF tables) :cite:p:`MZG02,Cha98`"""
PRESSURE_REFERENCE: float = 1.0
"""Standard state pressure in bar"""
GAS_STATE: str = "g"
"""Suffix to identify gases as per JANAF convention for the state of aggregation"""

# Initial solution guess
INITIAL_LOG_NUMBER_DENSITY: float = 50.0
"""Initial log number density

Empiricially determined. This value is mid-range for Earth-like planets.
"""
INITIAL_LOG_STABILITY: float = -30.0
"""Initial log stability.

Empirically determined.
"""

# Maximum x for which exp(x) is finite in 64-bit precision (to prevent overflow)
max_exp_input = np.log(np.finfo(np.float64).max)
# Minimum x for which exp(x) is non-zero in 64-bit precision
min_exp_input = np.log(jnp.finfo(np.float64).tiny)

# Lower and upper bounds on the hypercube which contains the root
LOG_NUMBER_DENSITY_LOWER: float = -170.0
"""Lower log number density for a species

For a gas species this corresponds to 3.17E-77 bar and 3.16E-78 bar at 3000 K and 298 K,
respectively.
"""
LOG_NUMBER_DENSITY_UPPER: float = 80.0
"""Upper log number density for a species

For a gas species this corresponds to 2294896 GPa and 227960 GPa at 3000 K and 298 K,
respectively. However, the choice of this upper limit is actually motivated by condensed species.
"""
LOG_STABILITY_LOWER: float = -700.0  # basically the same as min_exp_input
"""Lower stability for a species

Derived to ensure that the exponential function exp(x) does not underflow to zero
"""
LOG_STABILITY_UPPER: float = 35.0
"""Upper stability for a species

Empirically determined.
"""
TAU_MAX: float = 1.0e-3
"""Maximum tau scaling factor for species stability when using the tau cascade solver"""
TAU: float = 1.0e-25
"""Desired (i.e. final/minimium) tau scaling factor for species stability :cite:p:`LKK16`.

Tau effectively controls the minimum non-zero number density of unstable species. Formally, it
defines the number density of an unstable pure condensate with an activity of 1/e, which
corresponds to a log stability of zero.

This value is typically appropriate for condensate stability only, but if you additionally apply 
stability criteria to gas species you should reduce this value, maybe as low as 1e-60 to 1e-72 if
you want to ensure you do not truncated O2 at low temperatures. Hence you can override this default
using an argument to :class:`atmodeller.classes.InteriorAtmosphere`.
"""
TAU_NUM: int = 2
"""Number of tau values to solve between TAU_MAX and TAU (inclusive) for the tau cascade solver

Empirically determined. Basically, once a solution has been found for TAU_MAX the solver can
immediately proceed to TAU. This usually solves within a few steps on the first attempt.
"""

# Create the package logger.
# https://docs.python.org/3/howto/logging.html#library-config
logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def complex_formatter() -> logging.Formatter:
    """Complex formatter"""
    fmt: str = "[%(asctime)s - %(name)-30s - %(lineno)03d - %(levelname)-9s - %(funcName)s()]"
    fmt += " - %(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def simple_formatter() -> logging.Formatter:
    """Simple formatter for logging

    Returns:
        Formatter for logging
    """
    fmt: str = "[%(asctime)s - %(name)-30s - %(levelname)-9s] - %(message)s"
    datefmt: str = "%H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def debug_logger() -> logging.Logger:
    """Sets up debug logging to the console.

    Returns:
        A logger
    """
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(logging.DEBUG)
    package_logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    return package_logger


def debug_file_logger() -> logging.Logger:
    """Sets up info logging to the console and debug logging to a file.

    Returns:
        A logger
    """
    # Console logger
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(logging.DEBUG)
    package_logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(logging.INFO)
    package_logger.addHandler(console_handler)
    # File logger
    file_handler: logging.Handler = logging.FileHandler(f"{__package__}.log")
    file_formatter: logging.Formatter = complex_formatter()
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    package_logger.addHandler(file_handler)

    return package_logger


from atmodeller.classes import InteriorAtmosphere  # noqa: E402, F401
from atmodeller.containers import (  # noqa: E402, F401
    Planet,
    SolverParameters,
    Species,
    SpeciesCollection,
)
from atmodeller.utilities import (  # noqa: E402, F401
    bulk_silicate_earth_abundances,
    earth_oceans_to_hydrogen_mass,
)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for the EOS models from :cite:t:`ZD09`"""

from atmodeller.utilities import unit_conversion

RTOL: float = 1.0e-8
"""Relative tolerance"""
ATOL: float = 1.0e-8
"""Absolute tolerance"""
MODEL_SUFFIX: str = "zhang09"
"""Suffix of the :cite:t:`ZD09` models"""


def test_H2O_volume_low_TP(check_values) -> None:
    """Tests H2O volume at 1203.15 K and 950 MPa :cite:t:`ZD09{Table 6}`"""
    expected: float = 22.20343433408026
    expected *= unit_conversion.cm3_to_m3
    check_values.volume(
        1203.15,
        9500,  # 950 MPa
        check_values.get_eos_model("H2O", MODEL_SUFFIX),
        expected,
        rtol=RTOL,
        atol=ATOL,
    )


def test_H2O_volume_high_TP(check_values) -> None:
    """Tests H2O volume at 1873.15 K and 2500 MPa :cite:t:`ZD09{Table 6}`"""
    expected: float = 19.41089977577485
    expected *= unit_conversion.cm3_to_m3
    check_values.volume(
        1873.15,
        25000,  # 2500 MPa
        check_values.get_eos_model("H2O", MODEL_SUFFIX),
        expected,
        rtol=RTOL,
        atol=ATOL,
    )


def test_H2O_volume_high_TP2(check_values) -> None:
    """Tests H2O volume at 1373.15 K and 3500 MPa :cite:t:`ZD09{Table 6}`"""
    expected: float = 16.02290245403692
    expected *= unit_conversion.cm3_to_m3
    check_values.volume(
        1373.15,
        35000,  # 3500 MPa
        check_values.get_eos_model("H2O", MODEL_SUFFIX),
        expected,
        rtol=RTOL,
        atol=ATOL,
    )


# def test_volume_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("H2O", MODEL_SUFFIX)
#     check_values.check_broadcasting("volume", model)


# def test_fugacity_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("H2O", MODEL_SUFFIX)
#     check_values.check_broadcasting("fugacity", model)



#
# Copyright 2024 Dan J. Bower, Fabian L. Seidler
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Core classes and functions for solubility laws.

Units for temperature and pressure are K and bar, respectively.
"""

from abc import abstractmethod
from typing import Optional

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike

from atmodeller import override
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.thermodata._redox_buffers import IronWustiteBuffer
from atmodeller.utilities import power_law


class Solubility(eqx.Module):
    """Solubility interface

    :meth:`~Solubility.jax_concentration` is defined in order to allow arguments to be passed by
    position to lax.switch.
    """

    @abstractmethod
    def concentration(
        self,
        fugacity: ArrayLike,
        *,
        temperature: Optional[ArrayLike] = None,
        pressure: Optional[ArrayLike] = None,
        fO2: Optional[ArrayLike] = None,
    ) -> Array:
        """Concentration in ppmw

        Args:
            fugacity: Fugacity in bar
            temperature: Temperature in K. Defaults to None for not used.
            pressure: Pressure in bar. Defaults to None for not used.
            fO2: fO2 in bar. Defaults to None for not used.

        Returns:
            Concentration in ppmw
        """

    def jax_concentration(
        self, fugacity: ArrayLike, temperature: ArrayLike, pressure: ArrayLike, fO2: ArrayLike
    ) -> Array:
        """Wrapper to pass concentration arguments by position to use with JAX lax.switch

        Args:
            fugacity: Fugacity in bar
            temperature: Temperature in K
            pressure: Pressure in bar
            fO2: fO2 in bar

        Returns:
            Concentration in ppmw
        """
        return self.concentration(fugacity, temperature=temperature, pressure=pressure, fO2=fO2)


class NoSolubility(Solubility):
    """No solubility"""

    @override
    def concentration(
        self,
        fugacity: ArrayLike,
        *,
        temperature: Optional[ArrayLike] = None,
        pressure: Optional[ArrayLike] = None,
        fO2: Optional[ArrayLike] = None,
    ) -> Array:
        del fugacity
        del temperature
        del pressure
        del fO2

        return jnp.array(0.0)  # For JAX compatibility


class SolubilityPowerLaw(Solubility):
    """A solubility power law

    Args:
        constant: Constant
        exponent: Exponent
    """

    constant: float = eqx.field(converter=float)
    """Constant"""
    exponent: float = eqx.field(converter=float)
    """Exponent"""

    @override
    def concentration(self, fugacity: ArrayLike, *args, **kwargs) -> Array:
        del args
        del kwargs

        return power_law(fugacity, self.constant, self.exponent)


class SolubilityPowerLawLog10(Solubility):
    """A solubility power law with log10 coefficients

    Args:
        log10_constant: Log10 constant
        log10_exponent: Log10 exponent

    """

    log10_constant: float = eqx.field(converter=float)
    """Log10 constant"""
    log10_exponent: float = eqx.field(converter=float)
    """Log10 exponent"""

    @override
    def concentration(self, fugacity: ArrayLike, **kwargs) -> Array:
        del kwargs

        return jnp.power(10, (self.log10_constant + self.log10_exponent * jnp.log10(fugacity)))


def fO2_temperature_correction(
    fO2: ArrayLike,
    *,
    temperature: ArrayLike,
    pressure: ArrayLike,
    reference_temperature: ArrayLike,
) -> Array:
    """Applies a temperature correction to fO2.

    Some experimentally derived solubility laws operate on absolute fO2, which depends on
    temperature and pressure. A temperature correction has to be applied to maintain the same
    fO2 shift at arbitrary temperature.

    Args:
        fO2: Absolute oxygen fugacity at `temperature` in bar
        temperature: Temperature in K
        pressure: Absolute pressure in bar
        reference_temperature: Reference temperature, which is usually the temperature at which the
            experiment was performed.

    Returns:
        Adjusted fO2
    """
    iron_wustite_buffer: RedoxBufferProtocol = IronWustiteBuffer()
    logiw_fugacity_at_current_temp: ArrayLike = iron_wustite_buffer.log10_fugacity(
        temperature, pressure
    )
    fo2_shift: Array = jnp.log10(fO2) - logiw_fugacity_at_current_temp

    logiw_fugacity_at_reference_temp: ArrayLike = iron_wustite_buffer.log10_fugacity(
        reference_temperature, pressure
    )
    adjusted_fo2: Array = jnp.power(10, logiw_fugacity_at_reference_temp + fo2_shift)

    return adjusted_fo2



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for H-O systems"""

import logging
from typing import Mapping

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import __version__, debug_logger
from atmodeller._mytypes import NpFloat
from atmodeller.classes import InteriorAtmosphere
from atmodeller.containers import ConstantFugacityConstraint, Planet, Species, SpeciesCollection
from atmodeller.interfaces import FugacityConstraintProtocol, SolubilityProtocol
from atmodeller.output import Output
from atmodeller.solubility import get_solubility_models
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import earth_oceans_to_hydrogen_mass

logger: logging.Logger = debug_logger()
logger.setLevel(logging.DEBUG)

RTOL: float = 1.0e-6
"""Relative tolerance"""
ATOL: float = 1.0e-6
"""Absolute tolerance"""
TOLERANCE: float = 5.0e-2
"""Tolerance of log output to satisfy comparison with FactSage and FastChem"""

solubility_models: Mapping[str, SolubilityProtocol] = get_solubility_models()

H2O_g: Species = Species.create_gas("H2O", solubility=solubility_models["H2O_peridotite_sossi23"])
H2_g: Species = Species.create_gas("H2")
O2_g: Species = Species.create_gas("O2")
species: SpeciesCollection = SpeciesCollection((H2O_g, H2_g, O2_g))

gas_HO_system: InteriorAtmosphere = InteriorAtmosphere(species)


def test_version():
    """Test version."""
    assert __version__ == "0.8.2"


def test_H2O(helper) -> None:
    """Tests a single species (H2O)."""

    H2O_g: Species = Species.create_gas(
        "H2O", solubility=solubility_models["H2O_peridotite_sossi23"]
    )
    species: SpeciesCollection = SpeciesCollection((H2O_g,))
    planet: Planet = Planet()
    interior_atmosphere: InteriorAtmosphere = InteriorAtmosphere(species)

    oceans: ArrayLike = 2
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    mass_constraints: dict[str, ArrayLike] = {"H": h_kg}

    interior_atmosphere.solve(
        planet=planet,
        mass_constraints=mass_constraints,
    )
    output: Output = interior_atmosphere.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, float] = {
        "H2O_g": 1.0312913336898137,
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)


def test_H_O(helper) -> None:
    """Tests H2-H2O at the IW buffer by applying an oxygen abundance constraint."""

    species: SpeciesCollection = SpeciesCollection.create(("H2_g", "H2O_g", "O2_g"))
    planet: Planet = Planet()
    interior_atmosphere: InteriorAtmosphere = InteriorAtmosphere(species)

    oceans: ArrayLike = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    o_kg: ArrayLike = 6.25774e20
    mass_constraints: dict[str, ArrayLike] = {"H": h_kg, "O": o_kg}

    interior_atmosphere.solve(
        planet=planet,
        mass_constraints=mass_constraints,
    )
    output: Output = interior_atmosphere.output
    solution: dict[str, ArrayLike] = output.quick_look()

    fastchem_result: dict[str, float] = {
        "H2O_g": 76.45861543,
        "H2_g": 73.84378192,
        "O2_g": 8.91399329e-08,
    }

    assert helper.isclose(solution, fastchem_result, log=True, rtol=TOLERANCE, atol=TOLERANCE)


def test_H_fO2(helper) -> None:
    """Tests H2-H2O at the IW buffer with H2O solubility."""

    planet: Planet = Planet()
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(),
    }
    oceans: float = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    mass_constraints: dict[str, ArrayLike] = {"H": h_kg}

    gas_HO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = gas_HO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, float] = {
        "H2O_g": 0.2570800742364775,
        "H2_g": 0.2491511264610601,
        "O2_g": 8.838513516896038e-08,
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)


def test_H_fO2_fH2(helper) -> None:
    """Tests H2-H2O at the IW buffer with H2O solubility and mixed fugacity constraints."""

    planet: Planet = Planet()
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(np.array([-1, 0, 1])),
        "H2_g": ConstantFugacityConstraint(np.array([1.0e-8, 1.0e-7, 1.0e-6])),
    }

    gas_HO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
    )
    output: Output = gas_HO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, ArrayLike] = {
        "H2O_g": np.array([3.262913506271090e-09, 1.031823848794260e-07, 3.262913506271089e-06]),
        "H2_g": np.array([1.000000000000005e-08, 9.999999999999959e-08, 1.000000000000000e-06]),
        "O2_g": np.array([8.838513516896060e-09, 8.838513516896038e-08, 8.838513516896018e-07]),
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)


def test_H_fO2_batch_temperature(helper) -> None:
    """Tests H2-H2O at the IW buffer with H2O solubility for a range of surface temperatures."""

    # Number of surface temperatures is different to number of species to test array shapes work.
    surface_temperatures: NpFloat = np.array([1500, 2000, 2500, 3000])
    planet: Planet = Planet(surface_temperature=surface_temperatures)
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(),
        "H2_g": ConstantFugacityConstraint(np.nan),
    }
    oceans: float = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    mass_constraints: dict[str, ArrayLike] = {"H": h_kg}

    gas_HO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = gas_HO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, ArrayLike] = {
        "H2O_g": np.array(
            [
                2.566653037020448e-01,
                2.570800742364757e-01,
                2.572178041535549e-01,
                2.572746043480848e-01,
            ]
        ),
        "H2_g": np.array(
            [
                3.133632393608037e-01,
                2.491511264610584e-01,
                2.265704456625875e-01,
                2.199521409043987e-01,
            ]
        ),
        "O2_g": np.array(
            [
                2.394194493859141e-12,
                8.838513516896038e-08,
                4.544970468047975e-05,
                2.739422634823809e-03,
            ]
        ),
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)


def test_H_fO2_batch_fO2_shift(helper) -> None:
    """Tests H2-H2O at the IW buffer with H2O solubility for a range of fO2 shifts."""

    planet: Planet = Planet()
    # Set up a range of fO2 shifts
    num: int = 4
    fO2_shifts: NpFloat = np.linspace(-10, 10, num, dtype=np.float64)
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(fO2_shifts),
    }
    oceans: float = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    mass_constraints: dict[str, ArrayLike] = {"H": h_kg}

    gas_HO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = gas_HO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, ArrayLike] = {
        "H2O_g": np.array(
            [
                2.974916728388850e-04,
                1.609262626310640e-01,
                2.583020136413676e-01,
                2.585402668639946e-01,
            ]
        ),
        "H2_g": np.array(
            [
                2.883163373152490e01,
                7.239157580397396e00,
                5.393312233908341e-03,
                2.505662833497315e-06,
            ]
        ),
        "O2_g": np.array(
            [
                8.838513516896005e-18,
                4.102474564576031e-11,
                1.904200012911665e-04,
                8.838513516896137e02,
            ]
        ),
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)


def test_H_fO2_batch_H_mass(helper) -> None:
    """Tests H2-H2O at the IW buffer with H2O solubility for a range of H budgets."""

    planet: Planet = Planet()
    fugacity_constraints: dict[str, FugacityConstraintProtocol] = {
        "O2_g": IronWustiteBuffer(),
    }
    oceans: float = 1
    h_kg: ArrayLike = earth_oceans_to_hydrogen_mass(oceans)
    # Set up a range of H masses
    mass_constraints: dict[str, ArrayLike] = {"H": np.array([h_kg, 10 * h_kg, 100 * h_kg])}

    gas_HO_system.solve(
        planet=planet,
        fugacity_constraints=fugacity_constraints,
        mass_constraints=mass_constraints,
    )
    output: Output = gas_HO_system.output
    solution: dict[str, ArrayLike] = output.quick_look()

    target: dict[str, ArrayLike] = {
        "H2O_g": np.array([2.570800742364757e-01, 2.426110356931991e01, 1.610286613431932e03]),
        "H2_g": np.array([2.491511264610584e-01, 2.351283467393216e01, 1.560621626756960e03]),
        "O2_g": np.array([8.838513516896038e-08, 8.838513516896038e-08, 8.838513516896102e-08]),
    }

    assert helper.isclose(solution, target, rtol=RTOL, atol=ATOL)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""JAX-related functionality for solving the system of equations"""

from collections.abc import Callable
from typing import Any, cast

import equinox as eqx
import jax
import jax.numpy as jnp
import optimistix as optx
from jax import lax, random
from jax.scipy.special import logsumexp
from jaxtyping import Array, ArrayLike, Bool, Float, Integer, PRNGKeyArray, Shaped

from atmodeller.constants import AVOGADRO, BOLTZMANN_CONSTANT_BAR, GAS_CONSTANT
from atmodeller.containers import (
    FixedParameters,
    MassConstraints,
    Planet,
    SolverParameters,
    SpeciesCollection,
    TracedParameters,
)
from atmodeller.utilities import (
    get_log_number_density_from_log_pressure,
    safe_exp,
    to_hashable,
    unit_conversion,
)


# Since this is the core driver function for the solve it remains useful for debugging to see how
# many times recompilation is triggered
# @eqx.filter_jit
# @eqx.debug.assert_max_traces(max_traces=1)
def solve(
    solution_array: Float[Array, " sol_dim"],
    active_indices: Integer[Array, " res_dim"],
    tau: Float[Array, ""],
    traced_parameters: TracedParameters,
    solver_parameters: SolverParameters,
    fixed_parameters: FixedParameters,
    options: dict[str, Any],
) -> tuple[Float[Array, " sol_dim"], Bool[Array, ""], Integer[Array, ""]]:
    """Solves the system of non-linear equations

    Args:
        solution_array: Solution array
        active_indices: Indices of the residual array that are active
        tau: Tau parameter for species' stability
        traced_parameters: Traced parameters
        solver_parameters: Solver parameters
        fixed_parameters: Fixed parameters
        options: Options for root find

    Returns:
        The solution array, the status of the solver, number of steps
    """
    sol: optx.Solution = optx.root_find(
        objective_function,
        solver_parameters.get_solver_instance(),
        solution_array,
        args={
            "traced_parameters": traced_parameters,
            "active_indices": active_indices,
            "tau": tau,
            "fixed_parameters": fixed_parameters,
            "solver_parameters": solver_parameters,
        },
        throw=solver_parameters.throw,
        max_steps=solver_parameters.max_steps,
        options=options,
    )

    # jax.debug.print("Optimistix success. Number of steps = {out}", out=sol.stats["num_steps"])
    solver_steps: Integer[Array, ""] = sol.stats["num_steps"]

    # TODO: sol.results contains more information about the solution process, but it's wrapped up
    # in an enum-like object
    solver_status: Bool[Array, ""] = sol.result == optx.RESULTS.successful

    return sol.value, solver_status, solver_steps


def get_min_log_elemental_abundance_per_species(
    formula_matrix: Integer[Array, "el_dim species_dim"], mass_constraints: MassConstraints
) -> Float[Array, " species_dim"]:
    """For each species, find the elemental mass constraint with the lowest abundance.

    Args:
        formula_matrix: Formula matrix
        mass_constraints: Mass constraints

    Returns:
        A vector of the minimum log elemental abundance for each species
    """
    # Create the binary mask where formula_matrix != 0 (1 where element is present in species)
    mask: Integer[Array, "el_dim species_dim"] = (formula_matrix != 0).astype(jnp.int_)
    # jax.debug.print("formula_matrix = {out}", out=formula_matrix)
    # jax.debug.print("mask = {out}", out=mask)

    # log_abundance is a 1-D array, which cannot be transposed, so make a 2-D array
    log_abundance: Float[Array, "el_dim 1"] = jnp.atleast_2d(mass_constraints.log_abundance).T
    # jax.debug.print("log_abundance = {out}", out=log_abundance)

    # Element-wise multiplication with broadcasting
    masked_abundance: Float[Array, "el_dim species_dim"] = mask * log_abundance
    # jax.debug.print("masked_abundance = {out}", out=masked_abundance)
    masked_abundance = jnp.where(mask != 0, masked_abundance, jnp.nan)
    # jax.debug.print("masked_abundance = {out}", out=masked_abundance)

    # Find the minimum log abundance per species
    min_abundance_per_species: Float[Array, " species_dim"] = jnp.nanmin(masked_abundance, axis=0)
    # jax.debug.print("min_abundance_per_species = {out}", out=min_abundance_per_species)

    return min_abundance_per_species


def objective_function(
    solution: Float[Array, " sol_dim"], kwargs: dict
) -> Float[Array, " res_dim"]:
    """Objective function

    The order of the residual does make a difference to the solution process. More investigations
    are necessary, but justification for the current ordering is as follows:

        1. Fugacity constraints - fixed target, well conditioned
        2. Reaction constraints - log-linear, physics-based coupling
        3. Mass balance constraints - stiffer, depends on solubility
        4. Stability constraints - stiffer still

    Args:
        solution: Solution array for all species i.e. log number density and log stability
        kwargs: Dictionary of pytrees required to compute the residual

    Returns:
        Residual
    """
    # jax.debug.print("Starting new objective_function evaluation")
    tp: TracedParameters = kwargs["traced_parameters"]
    active_indices: Integer[Array, " res_dim"] = kwargs["active_indices"]
    fp: FixedParameters = kwargs["fixed_parameters"]
    tau: Float[Array, ""] = kwargs["tau"]
    planet: Planet = tp.planet
    temperature: Float[Array, ""] = planet.temperature

    log_number_density, log_stability = jnp.split(solution, 2)
    # jax.debug.print("log_number_density = {out}", out=log_number_density)
    # jax.debug.print("log_stability = {out}", out=log_stability)

    log_activity: Float[Array, " species_dim"] = get_log_activity(tp, fp, log_number_density)
    # jax.debug.print("log_activity = {out}", out=log_activity)

    # Atmosphere
    total_pressure: Float[Array, ""] = get_total_pressure(fp, log_number_density, temperature)
    # jax.debug.print("total_pressure = {out}", out=total_pressure)
    log_volume: Float[Array, ""] = get_atmosphere_log_volume(fp, log_number_density, planet)
    # jax.debug.print("log_volume = {out}", out=log_volume)

    # Based on the definition of the reaction constant we need to convert gas activities
    # (fugacities) from bar to effective number density, whilst keeping condensate activities
    # unmodified.
    log_activity_number_density: Float[Array, " species_dim"] = (
        get_log_number_density_from_log_pressure(log_activity, temperature)
    )
    log_activity_number_density = jnp.where(
        fp.gas_species_mask, log_activity_number_density, log_activity
    )
    # jax.debug.print("log_activity_number_density = {out}", out=log_activity_number_density)

    # Here would be where fugacity constraints could be imposed as hard constraints. Although this
    # would reduce the degrees of freedom, previous preliminary testing identified two challenges:
    #   1. The solver performance appears to degrade rather than improve. This could be because
    #       soft constraints are better behaved with gradient-based solution approaches(?)
    #   2. Imposing fugacity/activity would require back-computing pressure/number density, which
    #       would involve solving non-linear real gas EOS, potentially increasing the solve
    #       complexity and time.

    # Fugacity constraints residual (dimensionless, log-ratio of number densities)
    fugacity_residual = log_activity_number_density - tp.fugacity_constraints.log_number_density(
        temperature, total_pressure
    )
    # jax.debug.print("fugacity_residual = {out}", out=fugacity_residual)
    # jax.debug.print(
    #     "fugacity_residual min/max: {out}/{out2}",
    #     out=jnp.nanmin(fugacity_residual),
    #     out2=jnp.nanmax(fugacity_residual),
    # )
    # jax.debug.print(
    #     "fugacity_residual mean/std: {out}/{out2}",
    #     out=jnp.nanmean(fugacity_residual),
    #     out2=jnp.nanstd(fugacity_residual),
    # )

    # Reaction network residual
    # TODO: Is it possible to remove this if statement?
    if fp.reaction_matrix.size > 0:
        log_reaction_equilibrium_constant: Array = get_log_reaction_equilibrium_constant(
            fp, temperature
        )
        # jax.debug.print(
        #     "log_reaction_equilibrium_constant = {out}", out=log_reaction_equilibrium_constant
        # )
        reaction_residual: Array = (
            fp.reaction_matrix.dot(log_activity_number_density) - log_reaction_equilibrium_constant
        )
        # jax.debug.print("reaction_residual before stability = {out}", out=reaction_residual)
        reaction_stability_mask: Array = jnp.broadcast_to(
            fp.active_stability(), fp.reaction_matrix.shape
        )
        reaction_stability_matrix: Array = fp.reaction_matrix * reaction_stability_mask
        # jax.debug.print("reaction_stability_matrix = {out}", out=reaction_stability_matrix)

        # Dimensionless (log K residual)
        reaction_residual = reaction_residual - reaction_stability_matrix.dot(
            safe_exp(log_stability)
        )
        # jax.debug.print("reaction_residual after stability = {out}", out=reaction_residual)
        # jax.debug.print(
        #     "reaction_residual min/max: {out}/{out2}",
        #     out=jnp.nanmin(reaction_residual),
        #     out2=jnp.nanmax(reaction_residual),
        # )
        # jax.debug.print(
        #     "reaction_residual mean/std: {out}/{out2}",
        #     out=jnp.nanmean(reaction_residual),
        #     out2=jnp.nanstd(reaction_residual),
        # )

    else:
        reaction_residual = jnp.atleast_1d(jnp.array([]))

    # Elemental mass balance residual
    # Number density of elements in the gas or condensed phase
    element_density: Float[Array, " el_dim"] = get_element_density(
        fp.formula_matrix, log_number_density
    )
    # jax.debug.print("element_density = {out}", out=element_density)
    element_melt_density: Float[Array, " el_dim"] = get_element_density_in_melt(
        tp, fp, fp.formula_matrix, log_number_density, log_activity, log_volume
    )
    # jax.debug.print("element_melt_density = {out}", out=element_melt_density)

    # Relative mass error, computed in log-space for numerical stability
    element_density_total: Float[Array, " el_dim"] = element_density + element_melt_density
    log_element_density_total: Float[Array, " el_dim"] = jnp.log(element_density_total)
    # jax.debug.print("log_element_density_total = {out}", out=log_element_density_total)
    log_target_density: Float[Array, " el_dim"] = tp.mass_constraints.log_number_density(
        log_volume
    )
    # jax.debug.print("log_target_density = {out}", out=log_target_density)

    # Dimensionless (ratio error - 1)
    mass_residual: Float[Array, " el_dim"] = (
        safe_exp(log_element_density_total - log_target_density) - 1
    )
    # Log-space residual can perform better when close to the solution
    # mass_residual = log_element_density_total - log_target_density
    # jax.debug.print("mass_residual = {out}", out=mass_residual)
    # jax.debug.print(
    #     "mass_residual min/max: {out}/{out2}",
    #     out=jnp.nanmin(mass_residual),
    #     out2=jnp.nanmax(mass_residual),
    # )
    # jax.debug.print(
    #     "mass_residual mean/std: {out}/{out2}",
    #     out=jnp.nanmean(mass_residual),
    #     out2=jnp.nanstd(mass_residual),
    # )

    # Stability residual
    log_min_number_density: Float[Array, " species_dim"] = (
        get_min_log_elemental_abundance_per_species(fp.formula_matrix, tp.mass_constraints)
        - log_volume
        + jnp.log(tau)
    )
    # jax.debug.print("log_min_number_density = {out}", out=log_min_number_density)
    # Dimensionless (log-ratio)
    stability_residual: Float[Array, " species_dim"] = (
        log_number_density + log_stability - log_min_number_density
    )
    # jax.debug.print("stability_residual = {out}", out=stability_residual)
    # jax.debug.print(
    #     "stability_residual min/max: {out}/{out2}",
    #     out=jnp.nanmin(stability_residual),
    #     out2=jnp.nanmax(stability_residual),
    # )
    # jax.debug.print(
    #     "stability_residual mean/std: {out}/{out2}",
    #     out=jnp.nanmean(stability_residual),
    #     out2=jnp.nanstd(stability_residual),
    # )

    # NOTE: Order must be compatible with active_indices
    residual = jnp.concatenate(
        [fugacity_residual, reaction_residual, mass_residual, stability_residual]
    )
    # jax.debug.print("residual (with nans) = {out}", out=residual)

    residual = jnp.take(residual, indices=active_indices)
    # jax.debug.print("residual = {out}", out=residual)

    return residual


def get_atmosphere_log_molar_mass(
    fixed_parameters: FixedParameters, log_number_density: Float[Array, " species_dim"]
) -> Float[Array, ""]:
    """Gets log molar mass of the atmosphere

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density

    Returns:
        Log molar mass of the atmosphere
    """
    gas_log_number_density: Float[Array, " species_dim"] = get_gas_species_data(
        fixed_parameters, log_number_density
    )
    gas_molar_mass: Float[Array, " species_dim"] = get_gas_species_data(
        fixed_parameters, jnp.array(fixed_parameters.molar_masses)
    )
    molar_mass: Float[Array, ""] = logsumexp(gas_log_number_density, b=gas_molar_mass) - logsumexp(
        gas_log_number_density, b=fixed_parameters.gas_species_mask
    )
    # jax.debug.print("molar_mass = {out}", out=molar_mass)

    return molar_mass


def get_atmosphere_log_volume(
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
    planet: Planet,
) -> Float[Array, ""]:
    """Gets log volume of the atmosphere

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        planet: Planet

    Returns:
        Log volume of the atmosphere
    """
    log_volume: Float[Array, ""] = (
        jnp.log(GAS_CONSTANT)
        + jnp.log(planet.temperature)
        - get_atmosphere_log_molar_mass(fixed_parameters, log_number_density)
        + jnp.log(planet.surface_area)
        - jnp.log(planet.surface_gravity)
    )

    return log_volume


def get_total_pressure(
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
    temperature: Float[Array, ""],
) -> Float[Array, ""]:
    """Gets total pressure

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Total pressure
    """
    gas_species_mask: Bool[Array, " species_dim"] = fixed_parameters.gas_species_mask
    pressure: Float[Array, " species_dim"] = get_pressure_from_log_number_density(
        log_number_density, temperature
    )
    gas_pressure: Float[Array, " species_dim"] = pressure * gas_species_mask
    # jax.debug.print("gas_pressure = {out}", out=gas_pressure)

    return jnp.sum(gas_pressure)


def get_element_density(
    formula_matrix: Integer[Array, "el_dim species_dim"],
    log_number_density: Float[Array, " species_dim"],
) -> Array:
    """Number density of elements in the gas or condensed phase

    Args:
        formula_matrix: Formula matrix
        log_number_density: Log number density

    Returns:
        Number density of elements in the gas or condensed phase
    """
    element_density: Float[Array, " el_dim"] = formula_matrix @ safe_exp(log_number_density)

    return element_density


def get_element_density_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    formula_matrix: Integer[Array, "el_dim species_dim"],
    log_number_density: Float[Array, " species_dim"],
    log_activity: Float[Array, " species_dim"],
    log_volume: Float[Array, ""],
) -> Float[Array, " species_dim"]:
    """Gets the number density of elements dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        formula_matrix: Formula matrix
        log_number_density: Log number density
        log_activity: Log activity
        log_volume: Log volume of the atmosphere

    Returns:
        Number density of elements dissolved in melt
    """
    species_melt_density: Float[Array, " species_dim"] = get_species_density_in_melt(
        traced_parameters,
        fixed_parameters,
        log_number_density,
        log_activity,
        log_volume,
    )
    element_melt_density: Float[Array, " species_dim"] = formula_matrix.dot(species_melt_density)

    return element_melt_density


def get_gas_species_data(
    fixed_parameters: FixedParameters, some_array: ArrayLike
) -> Shaped[Array, " species_dim"]:
    """Masks the gas species data from an array

    Args:
        fixed_parameters: Fixed parameters
        some_array: Some array to mask the gas species data from

    Returns:
        An array with gas species data from `some_array` and condensate entries zeroed
    """
    gas_data: Shaped[Array, " species_dim"] = fixed_parameters.gas_species_mask * some_array

    return gas_data


def get_log_activity(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
) -> Float[Array, " species_dim"]:
    """Gets the log activity

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density

    Returns:
        Log activity
    """
    planet: Planet = traced_parameters.planet
    temperature: Float[Array, ""] = planet.temperature
    species: SpeciesCollection = fixed_parameters.species
    total_pressure: Float[Array, ""] = get_total_pressure(
        fixed_parameters, log_number_density, temperature
    )
    # jax.debug.print("total_pressure = {out}", out=total_pressure)

    activity_funcs: list[Callable] = [
        to_hashable(species_.activity.log_activity) for species_ in species
    ]

    def apply_activity(index: ArrayLike) -> Float[Array, ""]:
        return lax.switch(
            index,
            activity_funcs,
            temperature,
            total_pressure,
        )

    indices: Integer[Array, " species_dim"] = jnp.arange(len(species))
    vmap_activity: Callable = eqx.filter_vmap(apply_activity, in_axes=(0,))
    log_activity_pure_species: Float[Array, " species_dim"] = vmap_activity(indices)
    # jax.debug.print("log_activity_pure_species = {out}", out=log_activity_pure_species)
    log_activity: Float[Array, " species_dim"] = get_log_activity_ideal_mixing(
        fixed_parameters, log_number_density, log_activity_pure_species
    )
    # jax.debug.print("log_activity = {out}", out=log_activity)

    return log_activity


def get_log_activity_ideal_mixing(
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
    log_activity_pure_species: Float[Array, " species_dim"],
) -> Float[Array, " species_dim"]:
    """Gets the log activity of species in the atmosphere assuming an ideal mixture

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity_pure_species: Log activity of the pure species

    Returns:
        Log activity of the species assuming ideal mixing in the atmosphere
    """
    gas_species_mask: Bool[Array, " species_dim"] = fixed_parameters.gas_species_mask
    number_density: Float[Array, " species_dim"] = safe_exp(log_number_density)
    gas_species_number_density: Float[Array, " species_dim"] = gas_species_mask * number_density
    atmosphere_log_number_density: Float[Array, ""] = jnp.log(jnp.sum(gas_species_number_density))

    log_activity_gas_species: Float[Array, " species_dim"] = (
        log_activity_pure_species + log_number_density - atmosphere_log_number_density
    )
    log_activity: Float[Array, " species_dim"] = jnp.where(
        gas_species_mask, log_activity_gas_species, log_activity_pure_species
    )

    return log_activity


def get_log_pressure_from_log_number_density(
    log_number_density: Float[Array, " species_dim"], temperature: Float[Array, ""]
) -> Float[Array, " species_dim"]:
    """Gets log pressure from log number density

    Args:
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Log pressure
    """
    log_pressure: Float[Array, " species_dim"] = (
        jnp.log(BOLTZMANN_CONSTANT_BAR) + jnp.log(temperature) + log_number_density
    )

    return log_pressure


def get_log_Kp(
    species: SpeciesCollection,
    reaction_matrix: Float[Array, "react_dim species_dim"],
    temperature: Float[Array, ""],
) -> Float[Array, " react_dim"]:
    """Gets log of the equilibrium constant in terms of partial pressures

    Args:
        species: Species
        reaction_matrix: Reaction matrix
        temperature: Temperature

    Returns:
        Log of the equilibrium constant in terms of partial pressures
    """
    gibbs_funcs: list[Callable] = [
        to_hashable(species_.data.get_gibbs_over_RT) for species_ in species
    ]

    def apply_gibbs(
        index: Integer[Array, ""], temperature: Float[Array, "..."]
    ) -> Float[Array, "..."]:
        return lax.switch(index, gibbs_funcs, temperature)

    indices: Integer[Array, " species_dim"] = jnp.arange(len(species))
    vmap_gibbs: Callable = eqx.filter_vmap(apply_gibbs, in_axes=(0, None))
    gibbs_values: Float[Array, "species_dim 1"] = vmap_gibbs(indices, temperature)
    # jax.debug.print("gibbs_values = {out}", out=gibbs_values)
    log_Kp: Float[Array, "react_dim 1"] = -1.0 * reaction_matrix @ gibbs_values

    return jnp.ravel(log_Kp)


def get_log_reaction_equilibrium_constant(
    fixed_parameters: FixedParameters, temperature: Float[Array, ""]
) -> Float[Array, " react_dim"]:
    """Gets the log equilibrium constant of the reactions

    Args:
        fixed_parameters: Fixed parameters
        temperature: Temperature

    Returns:
        Log equilibrium constant of the reactions
    """
    species: SpeciesCollection = fixed_parameters.species
    reaction_matrix: Float[Array, "react_dim species_dim"] = jnp.array(
        fixed_parameters.reaction_matrix
    )
    log_Kp: Float[Array, " react_dim"] = get_log_Kp(species, reaction_matrix, temperature)
    # jax.debug.print("lnKp = {out}", out=lnKp)
    delta_n: Float[Array, " react_dim"] = jnp.sum(
        reaction_matrix * fixed_parameters.gas_species_mask, axis=1
    )
    # jax.debug.print("delta_n = {out}", out=delta_n)
    log_Kc: Float[Array, " react_dim"] = log_Kp - delta_n * (
        jnp.log(BOLTZMANN_CONSTANT_BAR) + jnp.log(temperature)
    )
    # jax.debug.print("log10Kc = {out}", out=log_Kc)

    return log_Kc


def get_pressure_from_log_number_density(
    log_number_density: Float[Array, " species_dim"], temperature: Float[Array, ""]
) -> Float[Array, " species_dim"]:
    """Gets pressure from log number density

    Args:
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Pressure
    """
    return safe_exp(get_log_pressure_from_log_number_density(log_number_density, temperature))


def get_species_density_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
    log_activity: Float[Array, " species_dim"],
    log_volume: Float[Array, ""],
) -> Float[Array, " species_dim"]:
    """Gets the number density of species dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity: Log activity
        log_volume: Log volume of the atmosphere

    Returns:
        Number density of species dissolved in melt
    """
    molar_masses: Float[Array, " species_dim"] = jnp.array(fixed_parameters.molar_masses)
    melt_mass: Float[Array, ""] = traced_parameters.planet.melt_mass

    ppmw: Float[Array, " species_dim"] = get_species_ppmw_in_melt(
        traced_parameters, fixed_parameters, log_number_density, log_activity
    )

    species_melt_density: Float[Array, " species_dim"] = (
        ppmw
        * unit_conversion.ppm_to_fraction
        * AVOGADRO
        * melt_mass
        / (molar_masses * safe_exp(log_volume))
    )
    # jax.debug.print("species_melt_density = {out}", out=species_melt_density)

    return species_melt_density


def get_species_ppmw_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Float[Array, " species_dim"],
    log_activity: Float[Array, " species_dim"],
) -> Float[Array, " species_dim"]:
    """Gets the ppmw of species dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity: Log activity

    Returns:
        ppmw of species dissolved in melt
    """
    species: SpeciesCollection = fixed_parameters.species
    diatomic_oxygen_index: Integer[Array, ""] = jnp.array(fixed_parameters.diatomic_oxygen_index)
    temperature: Float[Array, ""] = traced_parameters.planet.temperature

    fugacity: Float[Array, " species_dim"] = safe_exp(log_activity)
    total_pressure: Float[Array, ""] = get_total_pressure(
        fixed_parameters, log_number_density, temperature
    )
    diatomic_oxygen_fugacity: Float[Array, ""] = jnp.take(fugacity, diatomic_oxygen_index)

    # NOTE: All solubility formulations must return a JAX array to allow vmap
    solubility_funcs: list[Callable] = [
        to_hashable(species_.solubility.jax_concentration) for species_ in species
    ]

    def apply_solubility(
        index: Integer[Array, ""], fugacity: Float[Array, ""]
    ) -> Float[Array, ""]:
        return lax.switch(
            index,
            solubility_funcs,
            fugacity,
            temperature,
            total_pressure,
            diatomic_oxygen_fugacity,
        )

    indices: Integer[Array, " species_dim"] = jnp.arange(len(species))
    vmap_solubility: Callable = eqx.filter_vmap(apply_solubility, in_axes=(0, 0))
    species_ppmw: Float[Array, " species_dim"] = vmap_solubility(indices, fugacity)
    # jax.debug.print("ppmw = {out}", out=ppmw)

    return species_ppmw


@eqx.filter_jit
# Useful for optimising how many times JAX compiles the solve function
# @eqx.debug.assert_max_traces(max_traces=1)
def repeat_solver(
    solver_vmap_fn: Callable,
    active_indices: Integer[Array, " res_dim"],
    tau: Float[Array, "..."],
    solution: Float[Array, "batch_dim sol_dim"],
    traced_parameters: TracedParameters,
    solver_parameters: SolverParameters,
    key: PRNGKeyArray,
) -> tuple[
    Float[Array, "batch_dim sol_dim"],
    Bool[Array, " batch_dim"],
    Integer[Array, " batch_dim"],
    Integer[Array, " batch_dim"],
]:
    """Repeat solver that perturbs the initial solution for cases that fail and tries again

    Args:
        solver_vmap_fn: Vmapped solver function with pre-bound fixed configuration
        active_indices: Indices of the residual array that are active
        tau: Tau parameter for species' stability
        solution: Solution
        traced_parameters: Traced parameters
        solver_paramters: Solver parameters
        key: Random key

    Returns:
        A tuple with the state: (solution, solver_status, solver_steps, solver_attempts)
    """

    def body_fn(state: tuple[Array, ...]) -> tuple[Array, ...]:
        """Perform one iteration of the solver retry loop

        Args:
            state: Tuple containing:
                i: Current attempt index
                key: PRNG key for random number generation
                solution: Current solution array
                status: Boolean array indicating successful solutions
                steps: Step count
                success_attempt: Integer array recording iteration of success for each entry

        Returns:
            Updated state tuple
        """
        i, key, solution, status, steps, success_attempt = state

        failed_mask: Bool[Array, " batch_dim"] = ~status
        key, subkey = random.split(key)

        # Perturb the (initial) solution for cases that failed. Something more sophisticated could
        # be implemented, such as a regressor or neural network to inform failed cases based on
        # successful solves.
        perturb_shape: tuple[int, int] = (solution.shape[0], solution.shape[1])
        raw_perturb: Float[Array, "batch_dim sol_dim"] = random.uniform(
            subkey, shape=perturb_shape, minval=-1.0, maxval=1.0
        )
        perturbations: Float[Array, "batch_dim sol_dim"] = jnp.where(
            failed_mask[:, None],
            solver_parameters.multistart_perturbation * raw_perturb,
            jnp.zeros_like(solution),
        )
        new_initial_solution: Float[Array, "batch_dim sol_dim"] = solution + perturbations
        # jax.debug.print("new_initial_solution = {out}", out=new_initial_solution)

        new_solution, new_status, new_steps = solver_vmap_fn(
            new_initial_solution, active_indices, tau, traced_parameters, solver_parameters
        )

        # Determine which entries to update: previously failed, now succeeded
        update_mask: Bool[Array, " batch_dim"] = failed_mask & new_status
        updated_i: Integer[Array, "..."] = i + 1
        updated_solution: Float[Array, "batch_dim sol_dim"] = cast(
            Array, jnp.where(update_mask[:, None], new_solution, solution)
        )
        updated_status: Bool[Array, " batch_dim"] = status | new_status
        updated_steps: Integer[Array, " batch_dim"] = cast(
            Array, jnp.where(update_mask, new_steps, steps)
        )
        updated_success_attempt: Array = jnp.where(update_mask, updated_i, success_attempt)

        return (
            updated_i,
            key,
            updated_solution,
            updated_status,
            updated_steps,
            updated_success_attempt,
        )

    def cond_fn(state: tuple[Array, ...]) -> Bool[Array, "..."]:
        """Check if the solver should continue retrying

        Args:
            state: Tuple containing:
                i: Current attempt index
                _: Unused (PRNG key)
                _: Unused (solution)
                status: Boolean array indicating success of each solution
                _: Unused (steps)
                _: Unused (success_attempt)

        Returns:
            A boolean array indicating whether retries should continue (True if any solution
            failed and attempts are still available)
        """
        i, _, _, status, _, _ = state

        return jnp.logical_and(i < solver_parameters.multistart, jnp.any(~status))

    # Try first solution
    first_solution, first_solver_status, first_solver_steps = solver_vmap_fn(
        solution, active_indices, tau, traced_parameters, solver_parameters
    )
    # Failback solution
    solution = cast(Array, jnp.where(first_solver_status[:, None], first_solution, solution))

    initial_state: tuple[Array, ...] = (
        jnp.array(1),  # First attempt of the repeat_solver
        key,
        solution,
        first_solver_status,
        first_solver_steps,
        first_solver_status.astype(int),  # 1 for solved, otherwise 0
    )

    _, _, final_solution, final_status, final_steps, final_success_attempt = lax.while_loop(
        cond_fn, body_fn, initial_state
    )

    return final_solution, final_status, final_steps, final_success_attempt


def make_solve_tau_step(
    solver_vmap_fn: Callable,
    active_indices: Integer[Array, " res_dim"],
    traced_parameters: TracedParameters,
    solver_parameters: SolverParameters,
) -> Callable:
    """Wraps the repeat solver to call it for different tau values

    Args:
        solver_vmap_fn: Vmapped solver function with pre-bound fixed configuration
        active_indices: Indices of the residual array that are active
        traced_parameters: Traced parameters
        solver_parameters: Solver parameters

    Returns:
        Wrapped solver for a single tau value
    """

    @eqx.filter_jit
    # @eqx.debug.assert_max_traces(max_traces=1)
    def solve_tau_step(carry: tuple, tau: Float[Array, " batch_dim"]) -> tuple[tuple, tuple]:
        (key, solution) = carry
        key, subkey = jax.random.split(key)

        new_solution, new_status, new_steps, success_attempt = repeat_solver(
            solver_vmap_fn,
            active_indices,
            tau,
            solution,
            traced_parameters,
            solver_parameters,
            subkey,
        )

        new_carry: tuple[PRNGKeyArray, Float[Array, "batch_dim sol_dim"]] = (key, new_solution)

        # Output current solution etc for this tau step
        out: tuple[Array, ...] = (new_solution, new_status, new_steps, success_attempt)

        return new_carry, out

    return solve_tau_step



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests solubility laws for other species"""

import inspect
import logging

import numpy as np
from jaxtyping import ArrayLike

from atmodeller import debug_logger
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.solubility import get_solubility_models
from atmodeller.solubility.core import Solubility
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import unit_conversion

logger: logging.Logger = debug_logger()
logger.setLevel(logging.WARNING)

RTOL: float = 1.0e-8
"""Relative tolerance"""
ATOL: float = 1.0e-8
"""Absolute tolerance"""

LOG10_SHIFT: ArrayLike = 0
IW: RedoxBufferProtocol = IronWustiteBuffer(LOG10_SHIFT)

# Test a non-unity fugacity so the exponent is relevant for a power law solubility.
TEST_FUGACITY: ArrayLike = 2
TEST_TEMPERATURE: ArrayLike = 2000
TEST_PRESSURE: ArrayLike = 2  # bar
# Several models are calibrated in the low GPa range, so use this instead
TEST_PRESSURE_GPA: ArrayLike = 2 * unit_conversion.GPa_to_bar  # GPa
TEST_FO2: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE))
TEST_FO2_GPA: ArrayLike = np.exp(IW.log_fugacity(TEST_TEMPERATURE, TEST_PRESSURE_GPA))

logger.info("TEST_FUGACITY = %e bar", TEST_FUGACITY)
logger.info("TEST_TEMPERATURE = %e K", TEST_TEMPERATURE)
logger.info("TEST_PRESSURE = %e bar", TEST_PRESSURE)
logger.info("TEST_PRESSURE_GPA = %e bar", TEST_PRESSURE_GPA)
logger.info("TEST_FO2 = %e bar", TEST_FO2)
logger.info("TEST_FO2_GPA = %e bar", TEST_FO2_GPA)

solubility_models: dict[str, Solubility] = get_solubility_models()


def test_Cl2_ano_dio_for_thomas(check_values) -> None:
    """Tests Cl in silicate melts :cite:p:`TW21`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["Cl2_ano_dio_for_thomas21"]
    # target_concentration: ArrayLike = 1987252.8978466734
    target_concentration: ArrayLike = 1800
    test_fugacity_Cl2_ano_dio: ArrayLike = 1.66e-6
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_Cl2_ano_dio,
        TEST_TEMPERATURE,
        TEST_PRESSURE_GPA,
        TEST_FO2_GPA,
    )


def test_Cl2_basalt_thomas(check_values) -> None:
    """Tests Cl in silicate melts :cite:p:`TW21`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["Cl2_basalt_thomas21"]
    # target_concentration: ArrayLike = 1111006.1746003036
    target_concentration: ArrayLike = 16000
    test_fugacity_Cl2_basalt: ArrayLike = 4.24e-4
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_Cl2_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE_GPA,
        TEST_FO2_GPA,
    )


def test_He_basalt(check_values) -> None:
    """He in tholeittic basalt melt :cite:p:`JWB86`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["He_basalt_jambon86"]
    # target_concentration: ArrayLike = 0.20013
    target_concentration: ArrayLike = 0.1
    test_fugacity_He_basalt: ArrayLike = 1
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_He_basalt,
        TEST_TEMPERATURE,
        TEST_PRESSURE,
        TEST_FO2,
    )


def test_N2_basalt_bernadou(check_values) -> None:
    """Tests N2 in basaltic silicate melt :cite:p:`BGF21`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["N2_basalt_bernadou21"]
    # target_concentration: ArrayLike = 1.8621737654355521
    target_concentration: ArrayLike = 20.86
    test_fugacity_N2_basalt_bernadou: ArrayLike = 741
    test_totalP_N2_basalt_bernadou: ArrayLike = 800
    test_temperature_N2_basalt_bernadou: ArrayLike = 1473.15
    test_fO2_N2_basalt_bernadou: ArrayLike = 38.9045145
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_N2_basalt_bernadou,
        test_temperature_N2_basalt_bernadou,
        test_totalP_N2_basalt_bernadou,
        test_fO2_N2_basalt_bernadou,
    )


def test_N2_basalt_dasgupta(check_values) -> None:
    """Tests N2 in silicate melts :cite:p:`DFP22`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["N2_basalt_dasgupta22"]
    # target_concentration: ArrayLike = 2.280293304957063
    target_concentration: ArrayLike = 1000
    test_fugacity_N2_basalt_dasgupta: ArrayLike = 1550
    test_totalP_N2_basalt_dasgupta: ArrayLike = 1708.7
    test_temperature_N2_basalt_dasgupta: ArrayLike = 1773.15
    test_fO2_N2_basalt_dasgupta: ArrayLike = 1.8e-13
    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_N2_basalt_dasgupta,
        test_temperature_N2_basalt_dasgupta,
        test_totalP_N2_basalt_dasgupta,
        test_fO2_N2_basalt_dasgupta,
    )


def test_N2_basalt_libourel(check_values) -> None:
    """Tests N2 in basalt (tholeiitic) magmas :cite:p:`LMH03`"""

    function_name: str = inspect.currentframe().f_code.co_name  # type: ignore
    solubility_model: Solubility = solubility_models["N2_basalt_libourel03"]
    # target_concentration: ArrayLike = 0.12236470428806222
    target_concentration: ArrayLike = 446.7
    test_fugacity_N2_basalt_libourel: ArrayLike = 0.20
    test_temperature_N2_basalt_libourel: ArrayLike = 1698.15
    test_fO2_N2_basalt_libourel: ArrayLike = 6.31e-17
    test_totalP_N2_basalt_libourel: ArrayLike = 1
    # target_concentration: ArrayLike = 1000

    check_values.concentration(
        function_name,
        solubility_model,
        target_concentration,
        test_fugacity_N2_basalt_libourel,
        test_temperature_N2_basalt_libourel,
        test_totalP_N2_basalt_libourel,
        test_fO2_N2_basalt_libourel,
    )



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""My types"""

from typing import TypeAlias

import numpy as np
import numpy.typing as npt
import optimistix as optx

OptxSolver: TypeAlias = (
    optx.AbstractRootFinder | optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser
)

NpArray: TypeAlias = npt.NDArray
NpBool: TypeAlias = npt.NDArray[np.bool_]
NpFloat: TypeAlias = npt.NDArray[np.float64]
NpInt: TypeAlias = npt.NDArray[np.int_]
Scalar: TypeAlias = int | float



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Tests for the EOS models from :cite:t:`HWZ58`"""

from atmodeller.eos import RealGas
from atmodeller.utilities import unit_conversion

# Probably due to rounding of the model parameters in the paper, some compressibilities in the
# table in the paper don't quite match exactly with what we compute. Hence relax the tolerance.
RTOL: float = 1.0e-4
"""Relative tolerance"""
ATOL: float = 1.0e-4
"""Absolute tolerance"""

MODEL_SUFFIX: str = "beattie_holley58"
"""Suffix of the :cite:t:`HWZ58` models"""


def test_H2_low(check_values) -> None:
    """:cite:t:`HWZ58{Table II}`"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(300, pressure, model, 1.06217, rtol=RTOL, atol=ATOL)


def test_H2_high(check_values) -> None:
    """:cite:t:`HWZ58{Table II}`"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.26294, rtol=RTOL, atol=ATOL)


def test_H2_high_fugacity(check_values) -> None:
    """Tests that a fugacity can be calculated"""
    model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.fugacity(1000, pressure, model, 1301.672235770893, rtol=RTOL, atol=ATOL)


def test_N2_low(check_values) -> None:
    """:cite:t:`HWZ58{Table III}`"""
    model = check_values.get_eos_model("N2", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(300, pressure, model, 1.00464, rtol=RTOL, atol=ATOL)


def test_N2_high(check_values) -> None:
    """:cite:t:`HWZ58{Table III}`"""
    model: RealGas = check_values.get_eos_model("N2", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.36551, rtol=RTOL, atol=ATOL)


def test_O2_low(check_values) -> None:
    """:cite:t:`HWZ58{Table IV}`"""
    model: RealGas = check_values.get_eos_model("O2", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(300, pressure, model, 0.95454, rtol=RTOL, atol=ATOL)


def test_O2_high(check_values) -> None:
    """:cite:t:`HWZ58{Table IV}`"""
    model: RealGas = check_values.get_eos_model("O2", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.28897, rtol=RTOL, atol=ATOL)


def test_CO2_low(check_values) -> None:
    """:cite:t:`HWZ58{Table V}`"""
    model: RealGas = check_values.get_eos_model("CO2", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(400, pressure, model, 0.81853, rtol=RTOL, atol=ATOL)


def test_CO2_high(check_values) -> None:
    """:cite:t:`HWZ58{Table V}`"""
    model: RealGas = check_values.get_eos_model("CO2", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.07058, rtol=RTOL, atol=ATOL)


def test_NH3_low(check_values) -> None:
    """:cite:t:`HWZ58{Table VI}`"""
    model: RealGas = check_values.get_eos_model("NH3", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(400, pressure, model, 0.56165, rtol=RTOL, atol=ATOL)


def test_NH3_high(check_values) -> None:
    """:cite:t:`HWZ58{Table VI}`"""
    model: RealGas = check_values.get_eos_model("NH3", MODEL_SUFFIX)
    pressure: float = 500 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 0.93714, rtol=RTOL, atol=ATOL)


def test_CH4_low(check_values) -> None:
    """:cite:t:`HWZ58{Table VII}`"""
    model: RealGas = check_values.get_eos_model("CH4", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(300, pressure, model, 0.85583, rtol=RTOL, atol=ATOL)


def test_CH4_high(check_values) -> None:
    """:cite:t:`HWZ58{Table VII}`"""
    model: RealGas = check_values.get_eos_model("CH4", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.36201, rtol=RTOL, atol=ATOL)


def test_He_low(check_values) -> None:
    """:cite:t:`HWZ58{Table VIII}`"""
    model: RealGas = check_values.get_eos_model("He", MODEL_SUFFIX)
    pressure: float = 100 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(300, pressure, model, 1.05148, rtol=RTOL, atol=ATOL)


def test_He_high(check_values) -> None:
    """:cite:t:`HWZ58{Table VIII}`"""
    model: RealGas = check_values.get_eos_model("He", MODEL_SUFFIX)
    pressure: float = 1000 * unit_conversion.atmosphere_to_bar
    check_values.compressibility(1000, pressure, model, 1.14766, rtol=RTOL, atol=ATOL)


# def test_volume_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
#     check_values.check_broadcasting("volume", model)


# def test_fugacity_with_broadcasting(check_values) -> None:
#     """Tests volume with broadcasting"""
#     model: RealGas = check_values.get_eos_model("H2", MODEL_SUFFIX)
#     check_values.check_broadcasting("fugacity", model)



from isofunks import Fxuv
from constants import s2yr
import numpy as np

t=1e6, F0=1e2, t0 = 1e6, t_sat = 5e8, beta = -1.23, step_fn = False, F_final = 0, t_pms = 1e7, pms_factor = 1e2
time = t*s2yr
assert 0 < time < t_pms
F_pms0 = F0*pms_factor
s = (np.log10(F0) - np.log10(F_pms0)) / (np.log10(t_pms) - np.log10(t0))
assert Fxuv(t,F0) == F_pms0*(time/t0)**s

t=1e6, F0=1e2, t0 = 1e6, t_sat = 5e8, beta = -1.23, step_fn = False, F_final = 0, t_pms = 1e7, pms_factor = 1e2
time = t*s2yr
assert time < t_sat
assert Fxuv(t,F0) == F0

t=1e6, F0=1e2, t0 = 1e6, t_sat = 0, beta = -1.23, step_fn = False, F_final = 0, t_pms = 1e7, pms_factor = 1e2
time = t*s2yr
assert time >= t_sat
assert step_fn == False
assert Fxuv(t,F0) == F0*(time/t_sat)**beta

time=1e6, F0=1e2, t0 = 1e6, t_sat = 5e8, beta = -1.23, step_fn = True, F_final = 0, t_pms = 1e7, pms_factor = 1e2
time = t*s2yr
assert step_fn == True
assert Fxuv(t,F0) == F_final